"""Module for writing metadata to XML files."""

from collections import OrderedDict
from copy import copy
from typing import Any, Dict, Optional, Sequence, Union

from pysdmx.errors import Invalid
from pysdmx.io.xml.__tokens import (
    AGENCY_ID,
    AGENCY_SCHEME,
    AS_STATUS,
    ATT,
    ATT_REL,
    CL,
    CL_LOW,
    CLASS,
    CON,
    CON_ID,
    CONDITIONAL,
    CORE_REP,
    CS,
    CUSTOM_TYPE,
    CUSTOM_TYPE_SCHEME,
    DEPARTMENT,
    DFW,
    DIM,
    DIM_REF,
    DSD,
    DSD_COMPS,
    EMAIL,
    ENUM,
    ENUM_FORMAT,
    FAX,
    GROUP,
    GROUP_DIM,
    GROUPS_LOW,
    ID,
    LOCAL_REP,
    MANDATORY,
    MANDATORY_LOW,
    MEASURE,
    MEASURE_RELATIONSHIP,
    MSR,
    NAME,
    NAME_PER,
    NAME_PER_SCHEME,
    OPTIONAL_LOW,
    PACKAGE,
    PAR_ID,
    PAR_VER,
    POSITION,
    REF,
    ROLE,
    RULE,
    RULE_SCHEME,
    TELEPHONE,
    TEXT_FORMAT,
    TEXT_TYPE,
    TIME_DIM,
    TRANS_SCHEME,
    TRANSFORMATION,
    UDO,
    UDO_SCHEME,
    URI,
    URN,
    URN_LOW,
    USAGE,
    VALUE_ITEM,
    VALUE_LIST,
    VALUE_LIST_LOW,
    VALUE_LISTS,
    VERSION,
    VTL_MAPPING_SCHEME,
    VTLMAPPING,
)
from pysdmx.io.xml.__write_aux import (
    ABBR_COM,
    ABBR_MSG,
    ABBR_STR,
    MSG_CONTENT_PKG_21,
    MSG_CONTENT_PKG_30,
    __escape_xml,
    __to_lower_camel_case,
    add_indent,
)
from pysdmx.model import (
    AgencyScheme,
    Codelist,
    Concept,
    ConceptScheme,
    CustomType,
    CustomTypeScheme,
    DataType,
    Facets,
    Hierarchy,
    NamePersonalisation,
    NamePersonalisationScheme,
    Ruleset,
    RulesetScheme,
    Transformation,
    TransformationScheme,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
    VtlCodelistMapping,
    VtlConceptMapping,
    VtlDataflowMapping,
    VtlMappingScheme,
    VtlScheme,
)
from pysdmx.model.__base import (
    Agency,
    AnnotableArtefact,
    Contact,
    IdentifiableArtefact,
    Item,
    ItemReference,
    ItemScheme,
    MaintainableArtefact,
    NameableArtefact,
    Reference,
    VersionableArtefact,
)
from pysdmx.model.dataflow import (
    Component,
    Dataflow,
    DataStructureDefinition,
    Group,
    Role,
)
from pysdmx.util import (
    parse_item_urn,
    parse_short_urn,
    parse_urn,
)

ANNOTATION_WRITER = OrderedDict(
    {
        "title": "AnnotationTitle",
        "type": "AnnotationType",
        "text": "AnnotationText",
        "url": "AnnotationURL",
    }
)

ROLE_MAPPING = {
    Role.DIMENSION: DIM,
    Role.ATTRIBUTE: ATT,
    Role.MEASURE: MEASURE,
}

STR_TYPES = Union[
    ItemScheme,
    Codelist,
    ConceptScheme,
    DataStructureDefinition,
    Dataflow,
    CustomTypeScheme,
    VtlMappingScheme,
    NamePersonalisationScheme,
    RulesetScheme,
    UserDefinedOperatorScheme,
    TransformationScheme,
]

STR_DICT_TYPE_LIST_21 = {
    AgencyScheme: "OrganisationSchemes",
    Codelist: "Codelists",
    ConceptScheme: "Concepts",
    DataStructureDefinition: "DataStructures",
    Dataflow: "Dataflows",
    CustomTypeScheme: "CustomTypes",
    VtlMappingScheme: "VtlMappings",
    NamePersonalisationScheme: "NamePersonalisations",
    RulesetScheme: "Rulesets",
    UserDefinedOperatorScheme: "UserDefinedOperators",
    TransformationScheme: "Transformations",
}


STR_DICT_TYPE_LIST_30 = {
    AgencyScheme: "AgencySchemes",
    Codelist: "Codelists",
    ConceptScheme: "ConceptSchemes",
    DataStructureDefinition: "DataStructures",
    Dataflow: "Dataflows",
    CustomTypeScheme: "CustomTypeSchemes",
    VtlMappingScheme: "VtlMappingSchemes",
    NamePersonalisationScheme: "NamePersonalisationSchemes",
    RulesetScheme: "RulesetSchemes",
    UserDefinedOperatorScheme: "UserDefinedOperatorSchemes",
    TransformationScheme: "TransformationSchemes",
}


def __write_annotable(annotable: AnnotableArtefact, indent: str) -> str:
    """Writes the annotations to the XML file."""
    if len(annotable.annotations) == 0:
        return ""

    child1 = indent
    child2 = add_indent(child1)
    child3 = add_indent(child2)

    outfile = f"{child1}<{ABBR_COM}:Annotations>"
    for annotation in annotable.annotations:
        if annotation.id is None:
            outfile += f"{child2}<{ABBR_COM}:Annotation>"
        else:
            outfile += f"{child2}<{ABBR_COM}:Annotation id={annotation.id!r}>"
        outfile = outfile.replace("'", '"')

        for attr, label in ANNOTATION_WRITER.items():
            if getattr(annotation, attr, None) is not None:
                value = getattr(annotation, attr)
                value = __escape_xml(str(value))

                if attr == "text":
                    head_tag = f'{ABBR_COM}:{label} xml:lang="en"'

                else:
                    head_tag = f"{ABBR_COM}:{label}"

                outfile += f"{child3}<{head_tag}>{value}</{ABBR_COM}:{label}>"

        outfile += f"{child2}</{ABBR_COM}:Annotation>"
    outfile += f"{child1}</{ABBR_COM}:Annotations>"
    return outfile


def __write_identifiable(
    identifiable: IdentifiableArtefact,
    indent: str,
) -> Dict[str, Any]:
    """Writes the IdentifiableArtefact to the XML file."""
    attributes = ""

    attributes += f" id={identifiable.id!r}"

    if identifiable.uri is not None:
        attributes += f" uri={identifiable.uri!r}"

    if identifiable.urn is not None:
        attributes += f" urn={identifiable.urn!r}"

    outfile = {
        "Annotations": __write_annotable(identifiable, indent),
        "Attributes": attributes,
    }

    return outfile


def __write_nameable(
    nameable: NameableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the NameableArtefact to the XML file."""
    outfile = __write_identifiable(nameable, indent)
    attrs = ["Name", "Description"]

    for attr in attrs:
        value = getattr(nameable, attr.lower(), None)
        if value is not None:
            value = __escape_xml(str(value))
            outfile[attr] = [
                (
                    f"{indent}"
                    f'<{ABBR_COM}:{attr} xml:lang="en">'
                    f"{value}"
                    f"</{ABBR_COM}:{attr}>"
                )
            ]
        elif attr == "Name":
            raise Invalid(
                f"Name is required for NameableArtefact id= {nameable.id}"
            )

    return outfile


def __write_versionable(
    versionable: VersionableArtefact, indent: str, references_30: bool = False
) -> Dict[str, Any]:
    """Writes the VersionableArtefact to the XML file."""
    outfile = __write_nameable(versionable, add_indent(indent))

    if not (references_30 and isinstance(versionable, AgencyScheme)):
        outfile["Attributes"] += f" version={versionable.version!r}"

    if versionable.valid_from is not None:
        valid_from_str = versionable.valid_from.strftime("%Y-%m-%dT%H:%M:%S")
        outfile["Attributes"] += f" validFrom={valid_from_str!r}"

    if versionable.valid_to is not None:
        valid_to_str = versionable.valid_to.strftime("%Y-%m-%dT%H:%M:%S")
        outfile["Attributes"] += f" validTo={valid_to_str!r}"

    return outfile


def __write_maintainable(
    maintainable: MaintainableArtefact,
    indent: str,
    references_30: bool = False,
) -> Dict[str, Any]:
    """Writes the MaintainableArtefact to the XML file."""
    outfile = __write_versionable(maintainable, indent, references_30)

    outfile["Attributes"] += (
        f" isExternalReference="
        f"{str(maintainable.is_external_reference).lower()!r}"
    )
    if not references_30 and not (isinstance(maintainable, AgencyScheme)):
        outfile["Attributes"] += (
            f" isFinal={str(maintainable.is_final).lower()!r}"
        )

    if isinstance(maintainable.agency, str):
        outfile["Attributes"] += f" agencyID={maintainable.agency!r}"
    else:
        outfile["Attributes"] += f" agencyID={maintainable.agency.id!r}"

    return outfile


def __write_contact(contact: Contact, indent: str) -> str:
    """Writes the contact to the XML file."""

    def __item_to_str(item: Any, ns: str, tag: str) -> str:
        return (
            f"{add_indent(indent)}"
            f"<{ns}:{tag}>"
            f"{__escape_xml(str(item))}"
            f"</{ns}:{tag}>"
        )

    def __items_to_str(items: Any, ns: str, tag: str) -> str:
        return "".join([__item_to_str(item, ns, tag) for item in items])

    outfile = f"{indent}<{ABBR_STR}:Contact>"
    if contact.name is not None:
        outfile += __item_to_str(contact.name, ABBR_COM, NAME)
    if contact.department is not None:
        outfile += __item_to_str(contact.department, ABBR_STR, DEPARTMENT)
    if contact.role is not None:
        outfile += __item_to_str(contact.role, ABBR_STR, ROLE)
    if contact.telephones is not None:
        outfile += __items_to_str(contact.telephones, ABBR_STR, TELEPHONE)
    if contact.faxes is not None:
        outfile += __items_to_str(contact.faxes, ABBR_STR, FAX)
    if contact.uris is not None:
        outfile += __items_to_str(contact.uris, ABBR_STR, URI)
    if contact.emails is not None:
        outfile += __items_to_str(contact.emails, ABBR_STR, EMAIL)
    outfile += f"{indent}</{ABBR_STR}:Contact>"

    return outfile


def __write_item(
    item: Item, indent: str, scheme: str, references_30: bool = False
) -> str:
    """Writes the item to the XML file."""
    item_name = VALUE_ITEM if scheme == VALUE_LIST else type(item).__name__
    head = f"{ABBR_STR}:" + item_name

    data = __write_nameable(item, add_indent(indent))
    attributes = data["Attributes"].replace("'", '"')
    outfile = f"{indent}<{head}{attributes}>"
    outfile += __export_intern_data(data)
    if isinstance(item, Agency) and len(item.contacts) > 0:
        for contact in item.contacts:
            outfile += __write_contact(contact, add_indent(indent))
    if isinstance(item, Concept) and (
        item.codes is not None
        or item.facets is not None
        or item.dtype is not None
    ):
        outfile += f"{add_indent(indent)}<{ABBR_STR}:{CORE_REP}>"
        format_enum_text = TEXT_FORMAT
        if item.codes is not None:
            format_enum_text = ENUM_FORMAT
            outfile += __write_enumeration(
                item.codes, add_indent(indent), references_30
            )
        if item.facets is not None or item.dtype is not None:
            outfile += __write_text_format(
                item.dtype, item.facets, format_enum_text, add_indent(indent)
            )
        outfile += f"{add_indent(indent)}</{ABBR_STR}:{CORE_REP}>"
    outfile += f"{indent}</{head}>"
    return outfile


def __write_groups(
    groups: list[Group], indent: str, references_30: bool = False
) -> str:
    out_file = ""
    for group in groups:
        out_file += (
            f"{indent}<{ABBR_STR}:{GROUP} {URN_LOW}={group.urn!r}"
            f" {ID}={group.id!r}>"
        )
        for dimension in group.dimensions:
            if references_30:
                out_file += (
                    f"{add_indent(indent)}"
                    f"<{ABBR_STR}:{GROUP_DIM}>"
                    f"{add_indent(add_indent(indent))}"
                    f"<{ABBR_STR}:{DIM_REF}>{dimension}</{ABBR_STR}:{DIM_REF}>"
                    f"{add_indent(indent)}</{ABBR_STR}:{GROUP_DIM}>"
                )
            else:
                out_file += (
                    f"{add_indent(indent)}"
                    f"<{ABBR_STR}:{GROUP_DIM}>"
                    f"{add_indent(add_indent(indent))}"
                    f"<{ABBR_STR}:{DIM_REF}>"
                    f"{add_indent(add_indent(add_indent(indent)))}"
                    f"<{REF} {ID}={dimension!r}/>"
                    f"{add_indent(add_indent(indent))}"
                    f"</{ABBR_STR}:{DIM_REF}>"
                    f"{add_indent(add_indent(indent))}</{ABBR_STR}:{GROUP_DIM}>"
                )

        out_file += f"{indent}</{ABBR_STR}:{GROUP}>"
        out_file = out_file.replace("'", '"')
    return out_file


def __write_components(  # noqa: C901
    dsd: DataStructureDefinition, indent: str, references_30: bool = False
) -> str:
    """Writes the components to the XML file."""
    outfile = f"{indent}<{ABBR_STR}:{DSD_COMPS}>"

    components: Dict[str, Any] = {
        DIM: [],
        ATT: [],
        MEASURE: [],
    }
    out_group = ""
    groups = getattr(dsd, GROUPS_LOW, [])
    if groups is not None and len(groups) > 0:
        out_group = __write_groups(groups, add_indent(indent), references_30)

    for comp in dsd.components:
        if comp.role == Role.DIMENSION:
            components[DIM].append(comp)
        elif comp.role == Role.ATTRIBUTE:
            components[ATT].append(comp)
        else:
            components[MEASURE].append(comp)

    if not references_30 and len(components[MEASURE]) > 1:
        raise Invalid(
            title="Request cannot be fulfilled",
            description=f"SDMX-ML 2.1 does not support multiple measures. "
            f"Check the {dsd.short_urn}.",
            csi={
                "measures_found": components[MEASURE],
            },
        )

    position = 1
    for _, comps in components.items():
        if comps:
            role_name = ROLE_MAPPING[comps[0].role]
            if role_name == MEASURE:
                role_name = MSR
            outfile += f"{add_indent(indent)}<{ABBR_STR}:{role_name}List>"
            for comp in comps:
                outfile += __write_component(
                    comp,
                    position,
                    add_indent(add_indent(indent)),
                    components,
                    groups,
                    references_30,
                )
                position += 1
            outfile += f"{add_indent(indent)}</{ABBR_STR}:{role_name}List>"
            if role_name == DIM:
                outfile += out_group

    outfile += f"{indent}</{ABBR_STR}:{DSD_COMPS}>"
    return outfile


def __find_matching_group_id(
    att_rel: str, groups: list[Group]
) -> Optional[str]:
    comps_to_relate = att_rel.split(",") if "," in att_rel else [att_rel]
    for group in groups:
        if set(comps_to_relate) == set(group.dimensions):
            return group.id
    return None


def __comps_to_relate(
    att_rel: str,
    component_info: Dict[str, Any],
    references_30: bool = False,
) -> list[str]:
    if "," in att_rel:
        comps_to_relate = att_rel.split(",")
    elif att_rel == "O":
        if references_30:
            comps_to_relate = []
            for measure in component_info[MEASURE]:
                comps_to_relate.append(measure)
        else:
            comps_to_relate = [component_info[MEASURE][0].id]
    else:
        comps_to_relate = [att_rel]
    return comps_to_relate


def __write_attribute_relation(  # noqa: C901
    item: Component,
    indent: str,
    component_info: Dict[str, Any],
    groups: list[Group],
    references_30: bool = False,
) -> str:
    measure_relationship = ""
    outfile = f"{indent}<{ABBR_STR}:{ATT_REL}>"
    # At this point an attribute relation is always
    # present due to the post_init check in Component
    att_rel: str = item.attachment_level  # type: ignore[assignment]
    # Check if it is a list of Dimensions or it is related to the
    # primary measure
    group_id = None
    if groups is not None:
        group_id = __find_matching_group_id(att_rel, groups)
    comps_to_relate = __comps_to_relate(att_rel, component_info, references_30)
    dim_names = [comp.id for comp in component_info[DIM]]
    if references_30:
        if att_rel == "O":
            outfile += f"{add_indent(indent)}<{ABBR_STR}:Observation/>"
            measure_relationship += (
                f"{indent}<{ABBR_STR}:{MEASURE_RELATIONSHIP}>"
            )
            for comp_name in comps_to_relate:
                measure_relationship += (
                    f"{add_indent(indent)}<{ABBR_STR}:{MSR}>"
                    f"{comp_name.id}</{ABBR_STR}:{MSR}>"  # type: ignore[attr-defined]
                )
            measure_relationship += (
                f"{indent}</{ABBR_STR}:{MEASURE_RELATIONSHIP}>"
            )

        elif att_rel == "D":
            outfile += f"{add_indent(indent)}<{ABBR_STR}:Dataflow/>"
        elif group_id is not None:
            outfile += (
                f"{add_indent(indent)}<{ABBR_STR}:{GROUP}>"
                f"{group_id}</{ABBR_STR}:{GROUP}>"
            )
        else:
            for comp_name in comps_to_relate:
                outfile += (
                    f"{add_indent(indent)}<{ABBR_STR}:{DIM}>"
                    f"{comp_name}</{ABBR_STR}:{DIM}>"
                )

    else:
        if group_id is not None:
            outfile += (
                f"{add_indent(indent)}<{ABBR_STR}:{GROUP}>"
                f"{add_indent(add_indent(indent))}<{REF} "
                f"{ID}={group_id!r}/>"
                f"{add_indent(indent)}</{ABBR_STR}:{GROUP}>"
            )
        elif att_rel == "D":
            outfile += f"{add_indent(indent)}<{ABBR_STR}:None/>"

        else:
            for comp_name in comps_to_relate:
                role = (
                    Role.DIMENSION if comp_name in dim_names else Role.MEASURE
                )
                related_role = ROLE_MAPPING[role]
                outfile += f"{add_indent(indent)}<{ABBR_STR}:{related_role}>"
                outfile += (
                    f"{add_indent(add_indent(indent))}"
                    f"<{REF} {ID}={comp_name!r}/>"
                )
                outfile += f"{add_indent(indent)}</{ABBR_STR}:{related_role}>"

    outfile += f"{indent}</{ABBR_STR}:{ATT_REL}>"
    if measure_relationship:
        outfile += measure_relationship

    return outfile


def __write_component(
    item: Component,
    position: int,
    indent: str,
    component_info: Dict[str, Any],
    groups: list[Group],
    references_30: bool = False,
) -> str:
    """Writes the component to the XML file."""
    role_name = ROLE_MAPPING[item.role]
    if role_name == DIM and item.id == "TIME_PERIOD":
        role_name = TIME_DIM
    if references_30 and role_name == MEASURE:
        role_name = MSR

    head = f"{indent}<{ABBR_STR}:{role_name} "

    attributes = ""
    attribute_relation = ""
    if item.role == Role.ATTRIBUTE:
        if references_30:
            status = MANDATORY_LOW if item.required else OPTIONAL_LOW
            attributes += f"{USAGE}={status!r} "
        else:
            status = MANDATORY if item.required else CONDITIONAL
            attributes += f"{AS_STATUS}={status!r} "
        attribute_relation = __write_attribute_relation(
            item, add_indent(indent), component_info, groups, references_30
        )

    attributes += f"{ID}={item.id!r}"
    if item.role == Role.DIMENSION and not (
        role_name == TIME_DIM and references_30
    ):
        attributes += f" {POSITION}={str(position)!r}"

    if item.urn is not None:
        urn = item.urn
        if item.role == Role.MEASURE:
            # Handling URN to ensure roundtrip between SDMX 2.1 and 3.0
            if references_30:
                urn = urn.replace(".PrimaryMeasure", ".Measure")
            else:
                urn = urn.replace(".Measure", ".PrimaryMeasure")
        attributes += f" {URN.lower()}={urn!r}"

    attributes += ">"

    concept_identity = __write_concept_identity(
        item.concept, add_indent(indent), references_30
    )
    representation = __write_representation(
        item, add_indent(indent), references_30
    )

    outfile = head
    outfile += attributes
    outfile += concept_identity
    outfile += representation
    outfile += attribute_relation

    outfile += f"{indent}</{ABBR_STR}:{role_name}>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_concept_identity(
    identity: Union[Concept, ItemReference],
    indent: str,
    references_30: bool = False,
) -> str:
    if isinstance(identity, ItemReference):
        ref = identity
    else:
        ref = parse_item_urn(identity.urn)  # type: ignore[arg-type]

    outfile = f"{indent}<{ABBR_STR}:{CON_ID}>"
    if references_30:
        outfile += (
            f"urn:sdmx:org.sdmx.infomodel.conceptscheme.{ref.sdmx_type}={ref.agency}:"
            f"{ref.id}({ref.version}).{ref.item_id}"
            f"</{ABBR_STR}:{CON_ID}>"
        )
    else:
        outfile += f"{add_indent(indent)}<{REF} "
        outfile += f"{AGENCY_ID}={ref.agency!r} "
        outfile += f"{CLASS}={CON!r} "
        outfile += f"{ID}={ref.item_id!r} "
        outfile += f"{PAR_ID}={ref.id!r} "
        outfile += f"{PAR_VER}={ref.version!r} "
        outfile += f"{PACKAGE}={CS.lower()!r}/>"
        outfile += f"{indent}</{ABBR_STR}:{CON_ID}>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_representation(
    item: Component, indent: str, references_30: bool = False
) -> str:
    representation = ""
    local_representation = ""

    if item.local_codes is not None:
        local_representation += __write_enumeration(
            item.local_codes, indent, references_30
        )

    if item.local_facets is not None or item.local_dtype is not None:
        type_ = ENUM_FORMAT if item.local_codes is not None else TEXT_FORMAT
        local_representation += __write_text_format(
            item.local_dtype, item.local_facets, type_, indent
        )

    representation += f"{indent}<{ABBR_STR}:{LOCAL_REP}>"
    if len(local_representation) == 0:
        representation += f"{add_indent(indent)}<{ABBR_STR}:{TEXT_FORMAT}/>"
    representation += local_representation
    representation += f"{indent}</{ABBR_STR}:{LOCAL_REP}>"

    return representation


def __write_text_format(
    dtype: Optional[DataType],
    facets: Optional[Facets],
    type_: str,
    indent: str,
) -> str:
    """Writes the text format to the XML file."""
    outfile = f"{add_indent(indent)}<{ABBR_STR}:{type_}"
    if facets is not None:
        # Writing only facets not using default values
        active_facets = facets.__rich_repr__()
        for facet, value, *_ in active_facets:  # type: ignore[misc]
            facet = __to_lower_camel_case(facet)
            outfile += f' {facet}="{value}"'
    if dtype is not None:
        outfile += f" {TEXT_TYPE}={dtype.value!r}"
    outfile += "/>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_enumeration(
    codes: Union[Codelist, Hierarchy], indent: str, references_30: bool = False
) -> str:
    """Writes the enumeration to the XML file."""
    ref = parse_short_urn(codes.short_urn)
    outfile = f"{add_indent(indent)}<{ABBR_STR}:{ENUM}>"
    if references_30:
        outfile += (
            f"urn:sdmx:org.sdmx.infomodel.codelist.{ref.sdmx_type}={ref.agency}:{ref.id}({ref.version})"
            f"</{ABBR_STR}:{ENUM}>"
        )
    else:
        outfile += f"{add_indent(add_indent(indent))}<{REF} "
        outfile += f"{AGENCY_ID}={ref.agency!r} "
        outfile += f"{CLASS}={CL!r} "
        outfile += f"{ID}={ref.id!r} "
        outfile += f"{PACKAGE}={CL_LOW!r} "
        outfile += f"{VERSION}={ref.version!r}/>"
        outfile += f"{add_indent(indent)}</{ABBR_STR}:{ENUM}>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_structure(
    item: str, indent: str, references_30: bool = False
) -> str:
    """Writes the dataflow structure to the XML file."""
    ref = parse_short_urn(item)
    outfile = f"{indent}<{ABBR_STR}:Structure>"
    if references_30:
        outfile += (
            f"urn:sdmx:org.sdmx.infomodel.datastructure.{DSD}={ref.agency}:{ref.id}({ref.version})"
            f"</{ABBR_STR}:Structure>"
        )
    else:
        outfile += (
            f"{add_indent(indent)}<{REF} "
            f'{PACKAGE}="datastructure" '
            f"{AGENCY_ID}={ref.agency!r} "
            f"{ID}={ref.id!r} "
            f"{VERSION}={ref.version!r} "
            f"{CLASS}={DSD!r}/>"
        )
        outfile += f"{indent}</{ABBR_STR}:Structure>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_scheme(  # noqa: C901
    item_scheme: Any, indent: str, scheme: str, references_30: bool = False
) -> str:
    """Writes the scheme to the XML file."""
    if getattr(item_scheme, "sdmx_type", None) == "valuelist":
        scheme = VALUE_LIST
    label = f"{ABBR_STR}:{scheme}"
    components = ""
    data = __write_maintainable(item_scheme, indent, references_30)

    if scheme == DSD:
        components = __write_components(
            item_scheme, add_indent(indent), references_30
        )

    if scheme not in [
        DSD,
        DFW,
    ]:
        data["Attributes"] += (
            f" isPartial={str(item_scheme.is_partial).lower()!r}"
        )
    if scheme in [
        RULE_SCHEME,
        UDO_SCHEME,
        TRANS_SCHEME,
        VTL_MAPPING_SCHEME,
        CUSTOM_TYPE_SCHEME,
        NAME_PER_SCHEME,
    ]:
        data["Attributes"] += (
            f" {_write_vtl(item_scheme, indent, references_30)}"
        )

    outfile = ""

    attributes = data.get("Attributes") or ""
    attributes = attributes.replace("'", '"')

    outfile += f"{indent}<{label}{attributes}>"

    outfile += __export_intern_data(data)

    outfile += components

    if scheme == DFW:
        outfile += __write_structure(
            item_scheme.structure, add_indent(indent), references_30
        )

    if scheme not in [
        DSD,
        DFW,
        RULE_SCHEME,
        UDO_SCHEME,
        TRANS_SCHEME,
        VTL_MAPPING_SCHEME,
        CUSTOM_TYPE_SCHEME,
        NAME_PER_SCHEME,
    ]:
        for item in item_scheme.items:
            if (
                scheme == AGENCY_SCHEME
                and item.urn is not None
                and references_30
            ):
                agency_id = parse_short_urn(item_scheme.short_urn).agency
                item = copy(
                    item.__replace__(
                        urn=f"urn:sdmx:org.sdmx.infomodel.base.Agency={agency_id}:AGENCIES(1.0).{item.id}"
                    )
                )
            outfile += __write_item(
                item, add_indent(indent), scheme, references_30
            )
    if scheme in [
        RULE_SCHEME,
        UDO_SCHEME,
        TRANS_SCHEME,
        VTL_MAPPING_SCHEME,
        CUSTOM_TYPE_SCHEME,
        NAME_PER_SCHEME,
    ]:
        for item in item_scheme.items:
            outfile += _write_vtl(item, add_indent(indent), references_30)
        outfile += _write_vtl_references(
            item_scheme, add_indent(indent), references_30
        )
    outfile += f"{indent}</{label}>"

    return outfile


def __check_sdmx_type(
    package: Dict[str, Any], key: str, msg_content: Dict[str, Any]
) -> str:
    first_value = next(iter(package[key].values()), None)

    if (
        first_value is not None
        and getattr(first_value, "sdmx_type", None) == VALUE_LIST_LOW
    ):
        return VALUE_LISTS
    else:
        return msg_content[key]


def __write_metadata_element(
    package: Dict[str, Any],
    key: str,
    prettyprint: object,
    references_30: bool = False,
) -> str:
    """Writes the metadata element to the XML file.

    Args:
        package: The package to be written
        key: The key to be used
        prettyprint: Prettyprint or not
        references_30: Whether to use SDMX 3.0 references

    Returns:
        A string with the metadata element
    """
    outfile = ""
    nl = "\n" if prettyprint else ""
    child2 = "\t\t" if prettyprint else ""

    base_indent = f"{nl}{child2}"

    msg_content = MSG_CONTENT_PKG_30 if references_30 else MSG_CONTENT_PKG_21

    if key in package:
        scheme = __check_sdmx_type(package, key, msg_content)
        outfile += f"{base_indent}<{ABBR_STR}:{scheme}>"
        for element in package[key].values():
            item = (
                DSD
                if issubclass(element.__class__, DataStructureDefinition)
                else element.__class__.__name__
            )
            outfile += __write_scheme(
                element, add_indent(base_indent), item, references_30
            )

        outfile += f"{base_indent}</{ABBR_STR}:{scheme}>"

    return outfile


def __get_outfile(obj_: Dict[str, Any], key: str = "") -> str:
    """Generates an outfile from the object.

    Args:
        obj_: The object to be used
        key: The key to be used

    Returns:
        A string with the outfile

    """
    element = obj_.get(key) or []

    return "".join(element)


def __export_intern_data(data: Dict[str, Any]) -> str:
    """Export internal data (Annotations, Name, Description) on the XML file.

    Args:
        data: Information to be exported

    Returns:
        The XML string with the exported data
    """
    outfile = __get_outfile(data, "Annotations")
    outfile += __get_outfile(data, "Name")
    outfile += __get_outfile(data, "Description")

    return outfile


def __write_structures(
    content: Dict[str, Any], prettyprint: bool, references_30: bool = False
) -> str:
    """Writes the structures to the XML file.

    Args:
        content: The Message Content to be written
        prettyprint: Prettyprint or not
        references_30: Whether to use SDMX 3.0 references

    Returns:
        A string with the structures
    """
    nl = "\n" if prettyprint else ""
    child1 = "\t" if prettyprint else ""

    outfile = f"{nl}{child1}<{ABBR_MSG}:Structures>"
    msg_content = MSG_CONTENT_PKG_30 if references_30 else MSG_CONTENT_PKG_21
    for key in msg_content:
        outfile += __write_metadata_element(
            content, key, prettyprint, references_30
        )

    outfile += f"{nl}{child1}</{ABBR_MSG}:Structures>"

    # Replace &amp; with & in the outfile
    outfile = outfile.replace("& ", "&amp; ")

    return outfile


def _write_vtl(  # noqa: C901
    item_or_scheme: Union[Item, ItemScheme],
    indent: str,
    references_30: bool = False,
) -> str:  # noqa: C901
    """Writes the VTL attribute to the XML file for a single item.

    This function writes an item or an item scheme to the XML file,
    following the standard format.

    Args:
        item_or_scheme: The item or item scheme to be written
            Item: The item to be written
            ItemScheme: The item scheme to be written
        indent: The current indentation level
        references_30: Whether to use SDMX 3.0 references
    """
    outfile = ""

    if isinstance(item_or_scheme, Item):
        label = ""
        nameable = __write_nameable(item_or_scheme, add_indent(indent))
        attrib = nameable["Attributes"].replace("'", '"')
        data = __export_intern_data(nameable)

        if isinstance(item_or_scheme, Ruleset):
            label = f"{ABBR_STR}:{RULE}"
            data += f"{add_indent(indent)}<{ABBR_STR}:RulesetDefinition>"
            data += (
                f"{__escape_xml(item_or_scheme.ruleset_definition)}"
                f"</{ABBR_STR}:RulesetDefinition>"
            )
            attrib += (
                f" rulesetScope={item_or_scheme.ruleset_scope!r} "
                f"rulesetType={item_or_scheme.ruleset_type!r}"
            )

        if isinstance(item_or_scheme, Transformation):
            label = f"{ABBR_STR}:{TRANSFORMATION}"
            data += f"{add_indent(indent)}<{ABBR_STR}:Expression>"
            data += (
                f"{__escape_xml(item_or_scheme.expression)}"
                f"</{ABBR_STR}:Expression>"
            )
            data += f"{add_indent(indent)}<{ABBR_STR}:Result>"
            data += f"{item_or_scheme.result}</{ABBR_STR}:Result>"
            attrib += (
                f" isPersistent={str(item_or_scheme.is_persistent).lower()!r}"
            )

        if isinstance(item_or_scheme, UserDefinedOperator):
            label = f"{ABBR_STR}:{UDO}"
            data += f"{add_indent(indent)}<{ABBR_STR}:OperatorDefinition>"
            data += (
                f"{__escape_xml(item_or_scheme.operator_definition)}"
                f"</{ABBR_STR}:OperatorDefinition>"
            )
        if isinstance(item_or_scheme, VtlDataflowMapping):
            label = f"{ABBR_STR}:{VTLMAPPING}"
            attrib += f" alias={item_or_scheme.dataflow_alias!r}"
            data += f"{add_indent(indent)}<{ABBR_STR}:Dataflow>"
            reference = item_or_scheme.dataflow
            if references_30:
                data += (
                    f"urn:sdmx:org.sdmx.infomodel.datastructure.{DFW}={reference.agency}:{reference.id}"
                    f"({reference.version})</{ABBR_STR}:Dataflow>"
                )
            else:
                data += (
                    f"{indent}\t\t<{REF} package='datastructure' "
                    f"agencyID={reference.agency!r} id={reference.id!r} "
                    f"version={reference.version!r} class={DFW!r} />"
                    f"{add_indent(indent)}</{ABBR_STR}:Dataflow>"
                )
            if item_or_scheme.to_vtl_mapping_method is not None:
                to_vtl = item_or_scheme.to_vtl_mapping_method
                if len(to_vtl.to_vtl_sub_space) == 0:
                    data += (
                        f"{add_indent(indent)}<{ABBR_STR}:ToVtlMapping "
                        f"method='{to_vtl.method}' />"
                    )
                else:
                    data += (
                        f"{add_indent(indent)}<{ABBR_STR}:ToVtlMapping "
                        f"method='{to_vtl.method}'>"
                    )
                    indent_2 = add_indent(add_indent(indent))
                    data += f"{indent_2}<{ABBR_STR}:ToVtlSubSpace>"
                    for key in to_vtl.to_vtl_sub_space:
                        data += (
                            f"{add_indent(indent_2)}<{ABBR_STR}:Key>{key}"
                            f"</{ABBR_STR}:Key>"
                        )
                    data += f"{indent_2}</{ABBR_STR}:ToVtlSubSpace>"
                    data += f"{add_indent(indent)}</{ABBR_STR}:ToVtlMapping>"

            if item_or_scheme.from_vtl_mapping_method is not None:
                from_vtl = item_or_scheme.from_vtl_mapping_method
                if len(from_vtl.from_vtl_sub_space) == 0:
                    data += (
                        f"{add_indent(indent)}<{ABBR_STR}:FromVtlMapping "
                        f"method={from_vtl.method!r} />"
                        if from_vtl.method is not None
                        else f"method='{from_vtl.method}' />"
                    )
                else:
                    data += (
                        f"{add_indent(indent)}<{ABBR_STR}:FromVtlMapping "
                        f"method='{from_vtl.method}'>"
                    )
                    indent_2 = add_indent(add_indent(indent))
                    data += f"{indent_2}<{ABBR_STR}:FromVtlSuperSpace>"
                    for key in from_vtl.from_vtl_sub_space:
                        data += (
                            f"{add_indent(indent_2)}<{ABBR_STR}:Key>{key}"
                            f"</{ABBR_STR}:Key>"
                        )
                    data += f"{indent_2}</{ABBR_STR}:FromVtlSuperSpace>"
                    data += f"{add_indent(indent)}</{ABBR_STR}:FromVtlMapping>"

        if isinstance(item_or_scheme, VtlCodelistMapping):
            label = f"{ABBR_STR}:{VTLMAPPING}"
            data += f"{add_indent(indent)}<{ABBR_STR}:Codelist>"
            ref_codelist = (
                item_or_scheme.codelist
                if isinstance(item_or_scheme.codelist, Reference)
                else parse_urn(item_or_scheme.codelist)
                if isinstance(item_or_scheme.codelist, str)
                else parse_short_urn(item_or_scheme.codelist.short_urn)
            )
            if references_30:
                data += (
                    f"urn:sdmx:org.sdmx.infomodel.codelist.Codelist={ref_codelist.agency}:{ref_codelist.id}"
                    f"({ref_codelist.version})</{ABBR_STR}:Codelist>"
                )
            else:
                data += (
                    f"{indent}\t\t<{REF} package='codelist' "
                    f"agencyID={ref_codelist.agency!r} id={ref_codelist.id!r} "
                    f"version={ref_codelist.version!r} class={CL!r} />"
                    f"{add_indent(indent)}</{ABBR_STR}:Codelist>"
                )
            attrib += f" alias={item_or_scheme.codelist_alias!r}"

        if isinstance(item_or_scheme, VtlConceptMapping) and not isinstance(
            item_or_scheme.concept, Concept
        ):
            # TODO: Add handling for VtlConceptMapping
            #  when the Concept object is referenced
            label = f"{ABBR_STR}:{VTLMAPPING}"
            data += f"{add_indent(indent)}<{ABBR_STR}:Concept>"
            ref_concept = (
                parse_item_urn(item_or_scheme.concept)
                if isinstance(item_or_scheme.concept, str)
                else item_or_scheme.concept
            )
            if references_30:
                data += (
                    f"urn:sdmx:org.sdmx.infomodel.conceptscheme.Concept={ref_concept.agency}:{ref_concept.id}"
                    f"({ref_concept.version}).{ref_concept.item_id}</{ABBR_STR}:Concept>"
                )
            else:
                data += (
                    f"{indent}\t\t"
                    f"<{REF} maintainableParentID={ref_concept.id!r} "
                    f"package='conceptscheme' "
                    f"agencyID={ref_concept.agency!r} "
                    f"id={ref_concept.item_id!r} "
                    f"maintainableParentVersion={ref_concept.version!r} "
                    f"class={CON!r} />"
                    f"{add_indent(indent)}</{ABBR_STR}:Concept>"
                )
            attrib += f" alias={item_or_scheme.concept_alias!r}"

        if isinstance(item_or_scheme, CustomType):
            label = f"{ABBR_STR}:{CUSTOM_TYPE}"
            data += (
                f"{add_indent(indent)}<{ABBR_STR}:VtlScalarType>"
                f"{item_or_scheme.vtl_scalar_type}"
                f"</{ABBR_STR}:VtlScalarType>"
            )
            data += (
                f"{add_indent(indent)}<{ABBR_STR}:DataType>"
                f"{item_or_scheme.data_type}"
                f"</{ABBR_STR}:DataType>"
            )
            data += (
                (
                    f"{add_indent(indent)}<{ABBR_STR}:VtlLiteralFormat>"
                    f"{item_or_scheme.vtl_literal_format}"
                    f"</{ABBR_STR}:VtlLiteralFormat>"
                )
                if item_or_scheme.vtl_literal_format is not None
                else ""
            )
            data += (
                (
                    f"{add_indent(indent)}<{ABBR_STR}:OutputFormat>"
                    f"{item_or_scheme.output_format}"
                    f"</{ABBR_STR}:OutputFormat>"
                )
                if item_or_scheme.output_format is not None
                else ""
            )
            data += (
                (
                    f"{add_indent(indent)}<{ABBR_STR}:NullValue>"
                    f"{item_or_scheme.null_value}"
                    f"</{ABBR_STR}:NullValue>"
                )
                if item_or_scheme.null_value is not None
                else ""
            )

        if isinstance(item_or_scheme, NamePersonalisation):
            label = f"{ABBR_STR}:{NAME_PER}"
            attrib += f" vtlArtefact={item_or_scheme.vtl_artefact!r} "
            data += (
                f"{add_indent(indent)}<{ABBR_STR}:VtlDefaultName>"
                f"{item_or_scheme.vtl_default_name}"
                f"</{ABBR_STR}:VtlDefaultName>"
            )
            data += (
                f"{add_indent(indent)}<{ABBR_STR}:PersonalisedName>"
                f"{item_or_scheme.personalised_name}"
                f"</{ABBR_STR}:PersonalisedName>"
            )

        outfile += f"{indent}<{label}{attrib}>"
        outfile += data
        outfile += f"{indent}</{label}>"

    if isinstance(item_or_scheme, VtlScheme):
        outfile += f" vtlVersion={item_or_scheme.vtl_version!r}"

    outfile = outfile.replace("'", '"')

    return outfile


def _write_vtl_references(  # noqa: C901
    scheme: ItemScheme, indent: str, references_30: bool = False
) -> str:
    """Writes references to VTL elements to the XML file."""

    def process_references(
        references: Union[Any, Sequence[Any]], element_name: str
    ) -> str:
        """Process the references to VTL elements."""
        outreference = []
        if not isinstance(references, (list, tuple)):
            references = [references]

        for ref in references:
            if isinstance(ref, Reference):
                outreference.append(f"{indent}<{ABBR_STR}:{element_name}>")
                if references_30:
                    outreference.append(
                        f"urn:sdmx:org.sdmx.infomodel.{TRANSFORMATION.lower()}.{element_name}={ref.agency}:{ref.id}"
                        f"({ref.version})</{ABBR_STR}:{element_name}>"
                    )
                else:
                    outreference.append(
                        f"{add_indent(indent)}<{REF} "
                        f"{PACKAGE}={TRANSFORMATION.lower()!r} "
                        f"{AGENCY_ID}={ref.agency!r} "
                        f"{ID}={ref.id!r} "
                        f"{VERSION}={ref.version!r} "
                        f"{CLASS}={ref.sdmx_type!r}/>"
                        f"{indent}</{ABBR_STR}:{element_name}>"
                    )
            if isinstance(ref, ItemScheme):
                ref_to_use = parse_short_urn(ref.short_urn)
                outreference.append(f"{indent}<{ABBR_STR}:{element_name}>")
                if references_30:
                    outreference.append(
                        f"urn:sdmx:org.sdmx.infomodel.{TRANSFORMATION.lower()}.{element_name}={ref_to_use.agency}:"
                        f"{ref_to_use.id}({ref_to_use.version})</{ABBR_STR}:{element_name}>"
                    )
                else:
                    outreference.append(
                        f"{add_indent(indent)}<{REF} "
                        f"{PACKAGE}={TRANSFORMATION.lower()!r} "
                        f"{AGENCY_ID}={ref_to_use.agency!r} "
                        f"{ID}={ref_to_use.id!r} "
                        f"{VERSION}={ref_to_use.version!r} "
                        f"{CLASS}={ref_to_use.sdmx_type!r}/>"
                        f"{indent}</{ABBR_STR}:{element_name}>"
                    )

        return "".join(outreference)

    outfile = ""
    if isinstance(scheme, TransformationScheme):
        outfile += process_references(
            scheme.vtl_mapping_scheme, "VtlMappingScheme"
        )
        outfile += process_references(
            scheme.name_personalisation_scheme, "NamePersonalisationScheme"
        )
        outfile += process_references(
            scheme.custom_type_scheme, "CustomTypeScheme"
        )
        outfile += process_references(scheme.ruleset_schemes, "RulesetScheme")
        outfile += process_references(
            scheme.user_defined_operator_schemes, "UserDefinedOperatorScheme"
        )
    if isinstance(scheme, UserDefinedOperatorScheme):
        outfile += process_references(
            scheme.vtl_mapping_scheme, "VtlMappingScheme"
        )
        outfile += process_references(scheme.ruleset_schemes, "RulesetScheme")
    if isinstance(scheme, RulesetScheme):
        outfile += process_references(
            scheme.vtl_mapping_scheme, "VtlMappingScheme"
        )
    outfile = outfile.replace("'", '"')

    return outfile



"""Collection of SDMX-JSON schemas for organisations."""

from collections import defaultdict
from typing import Dict, Sequence, Set

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import ItemSchemeType
from pysdmx.io.json.sdmxjson2.messages.dataflow import JsonDataflow
from pysdmx.model import Agency, AgencyScheme, DataflowRef


class JsonAgencyScheme(ItemSchemeType, frozen=True):
    """SDMX-JSON payload for an agency scheme."""

    agencies: Sequence[Agency] = ()

    def __add_owner(
        self, owner: str, a: Agency, dfowners: Dict[str, Set[DataflowRef]]
    ) -> Agency:
        oid = f"{owner}.{a.id}" if owner != "SDMX" else a.id
        flows = list(dfowners[oid]) if dfowners else []
        return Agency(
            id=oid,
            name=a.name,
            description=a.description,
            contacts=a.contacts,
            dataflows=flows,
            annotations=[a.to_model() for a in self.annotations],
        )

    def to_model(self, dataflows: Sequence[JsonDataflow]) -> AgencyScheme:
        """Returns the requested list of agencies."""
        dfowners: Dict[str, Set[DataflowRef]] = defaultdict(set)
        if dataflows:
            for df in dataflows:
                dfref = DataflowRef(df.agency, df.id, df.version, df.name)
                dfowners[df.agency].add(dfref)
        agencies = [
            self.__add_owner(self.agency, a, dfowners) for a in self.agencies
        ]
        return AgencyScheme(
            description=self.description,
            agency=self.agency,
            items=agencies,
            annotations=[a.to_model() for a in self.annotations],
            is_external_reference=self.isExternalReference,
            is_partial=self.isPartial,
            valid_from=self.validFrom,
            valid_to=self.validTo,
        )


class JsonAgencySchemes(Struct, frozen=True):
    """SDMX-JSON payload for the list of agency schemes."""

    agencySchemes: Sequence[JsonAgencyScheme]
    dataflows: Sequence[JsonDataflow] = ()

    def to_model(self) -> Sequence[AgencyScheme]:
        """Returns the requested agency schemes."""
        return [a.to_model(self.dataflows) for a in self.agencySchemes]


class JsonAgencyMessage(Struct, frozen=True):
    """SDMX-JSON payload for /agencyscheme queries."""

    data: JsonAgencySchemes

    def to_model(self) -> Sequence[AgencyScheme]:
        """Returns the requested agency schemes."""
        return self.data.to_model()



"""Dataset module."""

from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, Optional, Union

from msgspec import Struct

from pysdmx.model import Schema


class ActionType(Enum):
    """Enumeration that defines the Dataset Action.

    Arguments:
        Append: Append data to an existing dataset.
        Replace: Replace the existing dataset with new data.
        Delete: Delete the data provided from the data source.
        Information: Provide information about the dataset
          without modifying it.

    """

    Append = "Append"
    Replace = "Replace"
    Delete = "Delete"
    Information = "Information"

    def __str__(self) -> str:
        """Return the action as a string."""
        return self.name.capitalize()

    def __repr__(self) -> str:
        """Action String representation."""
        return f"{self.__class__.__name__}.{self._name_}"


class SeriesInfo(Struct, frozen=True, repr_omit_defaults=True):
    """A group of related data, such as a time series, or a case series.

    Attributes:
        id: The unique ID for the series, such as a series key.
        name: The series's name (aka title), if any.
        obs_count: The number of observations available in the series.
        start_period: The oldest period for which data are available.
        end_period: The oldest period for which data are available.
        last_updated: When the series was last updated.
        is_active: Whether the series is still relevant or has been
            discontinued.
    """

    id: str
    name: Optional[str] = None
    obs_count: Optional[int] = None
    start_period: Optional[str] = None
    end_period: Optional[str] = None
    last_updated: Optional[datetime] = None
    is_active: bool = True

    def __str__(self) -> str:
        """Custom string representation without the class name."""
        processed_output = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            processed_output.append(f"{attr}: {value}")
        return f"{', '.join(processed_output)}"

    def __repr__(self) -> str:
        """Custom __repr__ that omits empty sequences."""
        attrs = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            attrs.append(f"{attr}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(attrs)})"


class Dataset(Struct, frozen=False, repr_omit_defaults=True, kw_only=True):
    """An organised collection of data.

    It includes metadata such as the structure of the dataset, attributes,
    action type, reporting periods and publication details.

    Args:
        structure: The structure referenced from a dataset,
          which can be a string (short_urn) or a Schema object.
        attributes: dictionary of attributes at dataset level, with its values.
        action: Defines the :class:`Action <pysdmx.model.dataset.ActionType>`
          of the dataset, default is ActionType.Information.
        reporting_begin: The start date for reporting, if applicable.
        reporting_end: The end date for reporting, if applicable.
        data_extraction_date: The date when the data was extracted.
        valid_from: The start date for the validity of the dataset.
        valid_to: The end date for the validity of the dataset.
        publication_year: The year of publication of the dataset.
        publication_period: The period of publication of the dataset.
        set_id: An optional identifier for the dataset.
    """

    structure: Union[str, Schema]
    attributes: Dict[str, Any] = {}
    action: ActionType = ActionType.Information
    reporting_begin: Optional[date] = None
    reporting_end: Optional[date] = None
    data_extraction_date: Optional[date] = None
    valid_from: Optional[date] = None
    valid_to: Optional[date] = None
    publication_year: Optional[date] = None
    publication_period: Optional[date] = None
    set_id: Optional[str] = None

    @property
    def short_urn(self) -> str:
        """Meaningful part of the URN.

        Returns:
            URN formatted string
        """
        if isinstance(self.structure, str):
            return self.structure
        else:
            return self.structure.short_urn

    def __str__(self) -> str:
        """Custom string representation without the class name."""
        processed_output = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            processed_output.append(f"{attr}: {value}")
        return f"{', '.join(processed_output)}"

    def __repr__(self) -> str:
        """Custom __repr__ that omits empty sequences."""
        attrs = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            attrs.append(f"{attr}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(attrs)})"



"""Parsers for reading metadata."""

from typing import Sequence, Union

from pysdmx.errors import Invalid
from pysdmx.io.xml.__parse_xml import parse_xml
from pysdmx.io.xml.__structure_aux_reader import StructureParser
from pysdmx.io.xml.__tokens import (
    STRUCTURE,
    STRUCTURES,
)
from pysdmx.model.__base import (
    ItemScheme,
)
from pysdmx.model.dataflow import (
    Dataflow,
    DataStructureDefinition,
)


def read(
    input_str: str,
    validate: bool = True,
) -> Sequence[Union[ItemScheme, DataStructureDefinition, Dataflow]]:
    """Reads an SDMX-ML 3.1 Structure data and returns the structures.

    Args:
        input_str: SDMX-ML structure message to read.
        validate: If True, the XML data will be validated against the XSD.

    Returns:
        dict: Dictionary with the parsed structures.
    """
    dict_info = parse_xml(input_str, validate)
    if STRUCTURE not in dict_info:
        raise Invalid("This SDMX document is not SDMX-ML 3.1 Structure.")
    return StructureParser(is_sdmx_30=True).format_structures(
        dict_info[STRUCTURE][STRUCTURES]
    )



"""Parsers for reading metadata."""

from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.io.xml.__tokens import (
    AGENCIES,
    AGENCY,
    AGENCY_ID,
    AGENCY_SCHEME,
    ALIAS_LOW,
    ANNOTATION,
    ANNOTATION_TEXT,
    ANNOTATION_TITLE,
    ANNOTATION_TYPE,
    ANNOTATION_URL,
    ANNOTATIONS,
    AS_STATUS,
    ATT,
    ATT_LIST,
    ATT_LVL,
    ATT_REL,
    ATTACH_GROUP,
    CL,
    CL_LOW,
    CLASS,
    CLS,
    CODE,
    CODES_LOW,
    COMPS,
    CON,
    CON_ID,
    CON_LOW,
    CON_ROLE,
    CON_SCHEMES,
    CONCEPTS,
    CONTACT,
    CORE_REP,
    CS,
    CUSTOM_TYPE,
    CUSTOM_TYPE_SCHEME,
    CUSTOM_TYPE_SCHEMES,
    CUSTOM_TYPES,
    DEPARTMENT,
    DESC,
    DFW,
    DFW_ALIAS_LOW,
    DFW_LOW,
    DFWS,
    DIM,
    DIM_LIST,
    DIM_REF,
    DSD,
    DSD_COMPS,
    DSDS,
    DTYPE,
    EMAIL,
    EMAILS,
    ENUM,
    ENUM_FORMAT,
    FACETS,
    FAX,
    FAXES,
    GROUP,
    GROUP_DIM,
    GROUPS_LOW,
    ID,
    IS_EXTERNAL_REF,
    IS_EXTERNAL_REF_LOW,
    IS_FINAL,
    IS_FINAL_LOW,
    IS_PARTIAL,
    IS_PARTIAL_LOW,
    LINK,
    LOCAL_CODES_LOW,
    LOCAL_DTYPE,
    LOCAL_FACETS_LOW,
    LOCAL_REP,
    MANDATORY,
    MANDATORY_LOW,
    ME_LIST,
    ME_REL,
    MEASURE,
    METADATA,
    MSR,
    NAME,
    NAME_PER,
    NAME_PER_SCHEME,
    NAME_PER_SCHEMES,
    NAME_PERS,
    OBSERVATION,
    ORGS,
    PAR_ID,
    PAR_VER,
    REF,
    REQUIRED,
    ROLE,
    RULE,
    RULE_SCHEME,
    RULE_SCHEMES,
    RULESETS,
    SER_URL,
    SER_URL_LOW,
    STR_URL,
    STR_URL_LOW,
    STRUCTURE,
    TELEPHONE,
    TELEPHONES,
    TEXT,
    TEXT_FORMAT,
    TEXT_TYPE,
    TIME_DIM,
    TITLE,
    TRANS_SCHEME,
    TRANS_SCHEMES,
    TRANSFORMATION,
    TRANSFORMATIONS,
    TYPE,
    UDO,
    UDO_SCHEME,
    UDO_SCHEMES,
    UDOS,
    URI,
    URIS,
    URL,
    URN,
    USAGE,
    VALID_FROM,
    VALID_FROM_LOW,
    VALID_TO,
    VALID_TO_LOW,
    VALUE_ITEM,
    VALUE_LIST,
    VALUE_LIST_LOW,
    VALUE_LISTS,
    VERSION,
    VTL_CL_MAPP,
    VTL_CON_MAPP,
    VTL_MAPPING_SCHEME,
    VTLMAPPING,
    VTLMAPPING_SCHEMES,
    VTLMAPPINGS,
)
from pysdmx.io.xml.utils import add_list
from pysdmx.model import (
    AgencyScheme,
    Code,
    Codelist,
    Concept,
    ConceptScheme,
    DataType,
    Facets,
    VtlCodelistMapping,
    VtlConceptMapping,
)
from pysdmx.model.__base import (
    Agency,
    Annotation,
    Contact,
    DataflowRef,
    Item,
    ItemReference,
    ItemScheme,
    Reference,
)
from pysdmx.model.dataflow import (
    Component,
    Components,
    Dataflow,
    DataStructureDefinition,
    Group,
    Role,
)
from pysdmx.model.vtl import (
    CustomType,
    CustomTypeScheme,
    FromVtlMapping,
    NamePersonalisation,
    NamePersonalisationScheme,
    Ruleset,
    RulesetScheme,
    ToVtlMapping,
    Transformation,
    TransformationScheme,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
    VtlDataflowMapping,
    VtlMappingScheme,
)
from pysdmx.util import find_by_urn, parse_urn

STRUCTURES_MAPPING = {
    CL: Codelist,
    VALUE_LIST: Codelist,
    AGENCY_SCHEME: AgencyScheme,
    CS: ConceptScheme,
    DFWS: Dataflow,
    DSDS: DataStructureDefinition,
    RULE_SCHEME: RulesetScheme,
    UDO_SCHEME: UserDefinedOperatorScheme,
    TRANS_SCHEME: TransformationScheme,
    VTL_MAPPING_SCHEME: VtlMappingScheme,
    NAME_PER_SCHEME: NamePersonalisationScheme,
    CUSTOM_TYPE_SCHEME: CustomTypeScheme,
}
ITEMS_CLASSES = {
    AGENCY: Agency,
    CODE: Code,
    VALUE_ITEM: Code,
    CON: Concept,
    RULE: Ruleset,
    UDO: UserDefinedOperator,
    TRANSFORMATION: Transformation,
    VTLMAPPING: VtlDataflowMapping,
    VTL_CL_MAPP: VtlCodelistMapping,
    VTL_CON_MAPP: VtlConceptMapping,
    NAME_PER: NamePersonalisation,
    CUSTOM_TYPE: CustomType,
}

COMP_TYPES = [DIM, ATT, MEASURE, MSR, GROUP_DIM]

ROLE_MAPPING = {
    DIM: Role.DIMENSION,
    ATT: Role.ATTRIBUTE,
    MEASURE: Role.MEASURE,
    MSR: Role.MEASURE,
}

FACETS_MAPPING = {
    "minLength": "min_length",
    "maxLength": "max_length",
    "minValue": "min_value",
    "maxValue": "max_value",
    "startValue": "start_value",
    "endValue": "end_value",
    "interval": "interval",
    "timeInterval": "time_interval",
    "decimals": "decimals",
    "pattern": "pattern",
    "startTime": "start_time",
    "endTime": "end_time",
    "isSequence": "is_sequence",
}


def _extract_text(element: Any) -> str:
    """Extracts the text from the element.

    Args:
        element: The element to extract the text from

    Returns:
        The text extracted from the element
    """
    if isinstance(element, list):
        aux = {}
        for language_element in element:
            if "lang" in language_element and language_element["lang"] == "en":
                aux = language_element
        if not aux:
            aux = element[0]
        element = aux
    if isinstance(element, dict) and "#text" in element:
        element = element["#text"]
    return element


def _format_lower_key(key: str, json_info: Dict[str, Any]) -> None:
    """Formats the key to lower case with underscores and returns it.

    Args:
        key: The key to be formatted
        json_info: The JSON information to be updated

    Returns:
        The formatted key in lower case

    """
    # Replaces the capital letters in the key with lower case,
    # adding an underscore before it if is not the first letter

    if key not in json_info:
        return
    formatted_key = key[0].lower() + "".join(
        "_" + c.lower() if c.isupper() else c for c in key[1:]
    )
    json_info[formatted_key] = json_info.pop(key)


class StructureParser(Struct):
    """StructureParser class for SDMX-ML."""

    agencies: Dict[str, AgencyScheme] = {}
    codelists: Dict[str, Codelist] = {}
    valuelists: Dict[str, Codelist] = {}
    concepts: Dict[str, ConceptScheme] = {}
    datastructures: Dict[str, DataStructureDefinition] = {}
    dataflows: Dict[str, Dataflow] = {}
    rulesets: Dict[str, RulesetScheme] = {}
    udos: Dict[str, UserDefinedOperatorScheme] = {}
    vtl_mappings: Dict[str, VtlMappingScheme] = {}
    name_personalisations: Dict[str, NamePersonalisationScheme] = {}
    custom_types: Dict[str, CustomTypeScheme] = {}
    transformations: Dict[str, TransformationScheme] = {}
    is_sdmx_30: bool = False

    def __format_contact(self, json_contact: Dict[str, Any]) -> Contact:
        """Creates a Contact object from a json_contact.

        Args:
            json_contact: The element to create the Contact object from

        Returns:
            Contact object created from the json_contact
        """
        self.__format_name_description(json_contact)

        xml_node_to_attribute = {
            NAME: NAME.lower(),
            DEPARTMENT: DEPARTMENT.lower(),
            ROLE: ROLE.lower(),
            URI: URIS,
            EMAIL: EMAILS,
            TELEPHONE: TELEPHONES,
            FAX: FAXES,
        }

        for k, v in xml_node_to_attribute.items():
            if k in json_contact:
                if k in [DEPARTMENT, ROLE]:
                    json_contact[v] = _extract_text(json_contact.pop(k))
                    continue
                field_info = add_list(json_contact.pop(k))
                for i, element in enumerate(field_info):
                    field_info[i] = _extract_text(element)
                json_contact[v] = field_info

        return Contact(**json_contact)

    @staticmethod
    def __format_annotations(item_elem: Any) -> Dict[str, Any]:
        """Formats the annotations in this element.

        Args:
            item_elem: The element to be formatted

        Returns:
            annotations formatted
        """
        if LINK in item_elem:
            del item_elem[LINK]
        if ANNOTATIONS not in item_elem:
            return item_elem
        annotations = []

        ann = item_elem[ANNOTATIONS]
        ann[ANNOTATION] = add_list(ann[ANNOTATION])
        for e in ann[ANNOTATION]:
            if ANNOTATION_TITLE in e:
                e[TITLE] = e.pop(ANNOTATION_TITLE)
            if ANNOTATION_TYPE in e:
                e[TYPE] = e.pop(ANNOTATION_TYPE)
            if ANNOTATION_TEXT in e:
                e[TEXT] = _extract_text(e[ANNOTATION_TEXT])
                del e[ANNOTATION_TEXT]
            if ANNOTATION_URL in e:
                e[URL] = e.pop(ANNOTATION_URL)

            annotations.append(Annotation(**e))

        item_elem[ANNOTATIONS.lower()] = annotations
        del item_elem[ANNOTATIONS]

        return item_elem

    @staticmethod
    def __format_name_description(element: Any) -> Dict[str, Any]:
        node = [NAME, DESC]
        for field in node:
            if field in element:
                element[field.lower()] = _extract_text(element[field])
                del element[field]
        return element

    @staticmethod
    def __format_facets(
        json_fac: Dict[str, Any], json_obj: Dict[str, Any]
    ) -> None:
        """Formats the facets from the JSON information.

        Args:
            json_fac: The element with the facets to be formatted
            json_obj: The element to store the formatted facets
        """
        if json_fac is None:
            return
        for key, _value in json_fac.items():
            if key == TEXT_TYPE and json_fac[TEXT_TYPE] in list(DataType):
                json_obj["dtype"] = DataType(json_fac[TEXT_TYPE])

            if key in FACETS_MAPPING:
                facet_kwargs = {
                    FACETS_MAPPING[k]: v
                    for k, v in json_fac.items()
                    if k in FACETS_MAPPING
                }
                json_obj[FACETS.lower()] = Facets(**facet_kwargs)

    @staticmethod
    def __format_validity(element: Dict[str, Any]) -> Dict[str, Any]:
        if VALID_FROM in element:
            element[VALID_FROM_LOW] = datetime.fromisoformat(
                element.pop(VALID_FROM)
            )
        if VALID_TO in element:
            element[VALID_TO_LOW] = datetime.fromisoformat(
                element.pop(VALID_TO)
            )
        return element

    @staticmethod
    def __format_urls(json_elem: Dict[str, Any]) -> Dict[str, Any]:
        """Formats the STR_URL and SER_URL keys in the element.

        Args:
            json_elem: The element to be formatted

        Returns:
            The json_elem with STR_URL and SER_URL keys formatted.
        """
        if STR_URL in json_elem:
            json_elem[STR_URL_LOW] = json_elem.pop(STR_URL)
        if SER_URL in json_elem:
            json_elem[SER_URL_LOW] = json_elem.pop(SER_URL)
        return json_elem

    def __format_agency(self, element: Dict[str, Any]) -> Dict[str, Any]:
        """Formats the AGENCY_ID key in the element to the maintainer.

        Args:
            element: The element with the Agency ID to be formatted

        Returns:
            element with the Agency ID formatted
        """
        element[AGENCY.lower()] = self.agencies.get(
            element[AGENCY_ID], element[AGENCY_ID]
        )
        del element[AGENCY_ID]
        return element

    def __format_orgs(self, json_orgs: Dict[str, Any]) -> Dict[str, Any]:
        orgs: Dict[str, Any] = {}
        json_list = add_list(json_orgs)
        for e in json_list:
            ag_sch = self.__format_scheme(
                e,
                AGENCY_SCHEME,
                AGENCY,
            )
            orgs = {**orgs, **ag_sch}
        return orgs

    def __format_representation(
        self, json_rep: Dict[str, Any], json_obj: Dict[str, Any]
    ) -> None:
        """Formats the representation in the JSON Representation."""
        if TEXT_FORMAT in json_rep:
            self.__format_facets(json_rep[TEXT_FORMAT], json_obj)

        if ENUM in json_rep and (
            len(self.codelists) > 0 or len(self.valuelists) > 0
        ):
            enum = json_rep[ENUM]
            if isinstance(enum, str):
                ref = parse_urn(enum)
            else:
                ref = enum.get(REF, enum)
            if isinstance(ref, dict) and "URN" in ref:
                codelist = find_by_urn(
                    list(self.codelists.values()), ref["URN"]
                )

            elif isinstance(ref, Reference):
                codelist = find_by_urn(
                    list(
                        self.codelists.values()
                        if ref.sdmx_type == CL
                        else self.valuelists.values()
                    ),
                    str(ref),
                )
            else:
                short_urn = str(
                    Reference(
                        sdmx_type=ref[CLASS],  # type: ignore[index]
                        agency=ref[AGENCY_ID],  # type: ignore[index]
                        id=ref[ID],  # type: ignore[index]
                        version=ref[VERSION],  # type: ignore[index]
                    )
                )
                codelist = self.codelists[short_urn]

            json_obj[CODES_LOW] = codelist
        if ENUM_FORMAT in json_rep:
            self.__format_facets(json_rep[ENUM_FORMAT], json_obj)

    def __format_local_rep(self, representation_info: Dict[str, Any]) -> None:
        rep: Dict[str, Any] = {}

        self.__format_representation(representation_info[LOCAL_REP], rep)
        del representation_info[LOCAL_REP]

        if CODES_LOW in rep:
            representation_info[LOCAL_CODES_LOW] = rep.pop(CODES_LOW)

        if DTYPE in rep:
            representation_info[LOCAL_DTYPE] = rep.pop(DTYPE)

        if FACETS.lower() in rep:
            representation_info[LOCAL_FACETS_LOW] = rep.pop(FACETS.lower())

    def __format_con_id(self, concept_ref: Dict[str, Any]) -> Dict[str, Any]:
        rep = {}
        if isinstance(concept_ref, str):
            item_reference = parse_urn(concept_ref)
            scheme_reference = Reference(
                sdmx_type=CS,
                agency=item_reference.agency,
                id=item_reference.id,
                version=item_reference.version,
            )
        else:
            item_reference = ItemReference(
                sdmx_type=concept_ref[CLASS],
                agency=concept_ref[AGENCY_ID],
                id=concept_ref[PAR_ID],
                version=concept_ref[PAR_VER],
                item_id=concept_ref[ID],
            )
            scheme_reference = Reference(
                sdmx_type=CS,
                agency=concept_ref[AGENCY_ID],
                id=concept_ref[PAR_ID],
                version=concept_ref[PAR_VER],
            )

        concept_scheme = self.concepts.get(str(scheme_reference))
        if concept_scheme is None:
            return {CON: item_reference}
        for con in concept_scheme.concepts:
            if isinstance(concept_ref, str):
                if con.id == item_reference.item_id:
                    rep[CON] = parse_urn(concept_ref)
                    break
            elif con.id == concept_ref[ID]:
                rep[CON] = con
                break
        if CON not in rep:
            return {CON: item_reference}
        return rep

    @staticmethod
    def __get_attachment_level(  # noqa: C901
        attribute: Dict[str, Any], element_info: Dict[str, Any]
    ) -> str:
        if DIM in attribute:
            dims = add_list(attribute[DIM])
            if dims and isinstance(dims[0], dict):
                dims = [dim[REF][ID] for dim in dims]
            att_level = ",".join(dims)
            # AttachmentGroup can only appear as sequence of the Dimension,
            # therefore we need to check first if a Dimension is present,
            # then the AttachmentGroup
            if ATTACH_GROUP in attribute:
                att_grp = add_list(attribute[ATTACH_GROUP])
                att_grp = [att[REF][ID] for att in att_grp]
                for grp in att_grp:
                    group_dims: List[str] = next(
                        (
                            g.dimensions
                            for g in element_info[GROUPS_LOW]
                            if g.id == grp
                        ),
                        [],
                    )
                    att_level += (
                        "," + ",".join(group_dims)
                        if len(group_dims) > 0
                        else ""
                    )
        elif GROUP in attribute:
            if REF in attribute[GROUP]:
                group_id = attribute[GROUP][REF][ID]
            else:
                group_id = attribute[GROUP]
            group_dimensions: List[str] = next(
                (
                    g.dimensions
                    for g in element_info[GROUPS_LOW]
                    if g.id == group_id
                ),
                [],
            )
            att_level = (
                ",".join(group_dimensions) if len(group_dimensions) > 0 else ""
            )
        elif OBSERVATION in attribute or MEASURE in attribute:
            att_level = "O"
        else:
            # For None (SDMX-2.1) or Dataflow (SDMX-3.0), attribute is
            # related to Dataset/Dataflow
            att_level = "D"

        return att_level

    def __format_vtl_references(
        self, json_elem: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Formats the references in the VTL element."""

        def extract_references(
            scheme: str,
            new_key: str,
            object_list: Dict[str, Any],
            as_list: bool = True,
        ) -> None:
            references = []
            if scheme in json_elem:
                scheme_entries = (
                    json_elem[scheme]
                    if isinstance(json_elem[scheme], list)
                    else [json_elem[scheme]]
                )
                for entry in scheme_entries:
                    if isinstance(entry, str):
                        entry_ref = parse_urn(entry)
                        ref_id = entry_ref.id
                        reference = entry_ref
                    else:
                        ref = entry[REF]
                        ref_id = ref[ID]
                        reference = Reference(
                            sdmx_type=ref[CLASS],
                            agency=ref[AGENCY_ID],
                            id=ref_id,
                            version=ref[VERSION],
                        )

                    matching_object = next(
                        (
                            obj
                            for obj in object_list.values()
                            if getattr(obj, ID, None) == ref_id
                        ),
                        None,
                    )

                    if matching_object:
                        references.append(matching_object)
                    else:
                        references.append(reference)

                if not as_list:
                    json_elem[new_key] = references[0]
                else:
                    json_elem[new_key] = references
                json_elem.pop(scheme)

        extract_references(RULE_SCHEME, "ruleset_schemes", self.rulesets)
        extract_references(
            UDO_SCHEME, "user_defined_operator_schemes", self.udos
        )
        extract_references(
            VTL_MAPPING_SCHEME,
            "vtl_mapping_scheme",
            self.vtl_mappings,
            as_list=False,
        )
        extract_references(
            NAME_PER_SCHEME,
            "name_personalisation_scheme",
            self.name_personalisations,
            as_list=False,
        )
        extract_references(
            CUSTOM_TYPE_SCHEME,
            "custom_type_scheme",
            self.custom_types,
            as_list=False,
        )
        return json_elem

    def __format_dataflow(
        self, json_rep: Dict[str, Any], json_obj: Dict[str, Any]
    ) -> None:
        json_obj[DFW_ALIAS_LOW] = json_obj.pop(ALIAS_LOW)
        if isinstance(json_rep, str):
            ref_aux = parse_urn(json_rep)
            dataflow_ref = {
                "agency": ref_aux.agency,
                "id": ref_aux.id,
                "version": ref_aux.version,
            }
        else:
            dataflow_ref = {
                "agency": json_rep[REF][AGENCY_ID],
                "id": json_rep[REF][ID],
                "version": json_rep[REF][VERSION],
            }
        json_obj[DFW_LOW] = DataflowRef(**dataflow_ref)
        if REF in json_rep:
            json_rep.pop(REF)
        json_obj.pop(DFW)
        if self.dataflows:
            for dataflow in self.dataflows.values():
                if dataflow.id == dataflow_ref[ID]:
                    json_obj[DFW_LOW] = dataflow

    def __format_component(
        self, comp: Dict[str, Any], role: Role, element_info: Dict[str, Any]
    ) -> Component:
        comp[ROLE.lower()] = role
        comp[REQUIRED] = True

        self.__format_local_rep(comp) if LOCAL_REP in comp else None

        if LINK in comp:
            del comp[LINK]
        if REF in comp[CON_ID]:
            concept_id = self.__format_con_id(comp[CON_ID][REF])
        else:
            concept_id = self.__format_con_id(comp[CON_ID])
        comp[CON_LOW] = concept_id.pop(CON)
        del comp[CON_ID]

        # Attribute Handling
        if ATT_REL in comp:
            comp[ATT_LVL] = self.__get_attachment_level(
                comp[ATT_REL], element_info
            )
            del comp[ATT_REL]

        if ME_REL in comp:
            del comp[ME_REL]

        if AS_STATUS in comp or USAGE in comp:
            status_key = AS_STATUS if AS_STATUS in comp else USAGE
            if (
                comp[status_key] != MANDATORY
                and comp[status_key] != MANDATORY_LOW
            ):
                comp[REQUIRED] = False
            del comp[status_key]

        if "position" in comp:
            del comp["position"]

        if ANNOTATIONS in comp:
            del comp[ANNOTATIONS]

        if CON_ROLE in comp:
            del comp[CON_ROLE]

        return Component(**comp)

    def __format_component_lists(
        self, element: Dict[str, Any], element_info: Dict[str, Any]
    ) -> List[Component]:
        comp_list = []

        if TIME_DIM in element:
            element[DIM] = add_list(element[DIM])
            element[DIM].append(element[TIME_DIM])
            del element[TIME_DIM]

        role_name = list(set(element.keys()).intersection(COMP_TYPES))[0]
        role = ROLE_MAPPING[role_name]
        element[role_name] = add_list(element[role_name])

        for comp in element[role_name]:
            formatted_comp = self.__format_component(
                comp,
                role,
                element_info,
            )
            comp_list.append(formatted_comp)

        return comp_list

    def __format_components(self, element: Dict[str, Any]) -> Dict[str, Any]:
        if DSD_COMPS in element:
            element[COMPS] = []
            comps = element[DSD_COMPS]

            for comp_list in [DIM_LIST, ME_LIST, ATT_LIST]:
                if comp_list in comps:
                    fmt_comps = self.__format_component_lists(
                        comps[comp_list], element
                    )
                    element[COMPS].extend(fmt_comps)

            element[COMPS] = Components(element[COMPS])
            del element[DSD_COMPS]

        return element

    def __format_vtl(self, json_vtl: Dict[str, Any]) -> Dict[str, Any]:
        # VTL Scheme Handling
        _format_lower_key("vtlVersion", json_vtl)
        # Transformation Scheme Handling
        if "isPersistent" in json_vtl:
            json_vtl["is_persistent"] = (
                json_vtl.pop("isPersistent").lower() == "true"
            )
        _format_lower_key("Expression", json_vtl)
        _format_lower_key("Result", json_vtl)

        # Ruleset Handling
        _format_lower_key("rulesetScope", json_vtl)
        _format_lower_key("rulesetType", json_vtl)
        _format_lower_key("RulesetDefinition", json_vtl)
        # User Defined Operator Handling
        _format_lower_key("OperatorDefinition", json_vtl)
        # Dataflow Mapping
        if "ToVtlMapping" in json_vtl:
            to_vtl = json_vtl.pop("ToVtlMapping")
            if "ToVtlSubSpace" in to_vtl:
                to_vtl["to_vtl_sub_space"] = add_list(
                    to_vtl["ToVtlSubSpace"]["Key"]
                )
                del to_vtl["ToVtlSubSpace"]
            json_vtl["to_vtl_mapping_method"] = ToVtlMapping(**to_vtl)
        if "FromVtlMapping" in json_vtl:
            from_vtl = json_vtl.pop("FromVtlMapping")
            if "FromVtlSuperSpace" in from_vtl:
                from_vtl["from_vtl_sub_space"] = add_list(
                    from_vtl["FromVtlSuperSpace"]["Key"]
                )
                del from_vtl["FromVtlSuperSpace"]
            json_vtl["from_vtl_mapping_method"] = FromVtlMapping(**from_vtl)
        # Codelist Mapping
        if CL in json_vtl:
            if isinstance(json_vtl[CL], str):
                cl_ref_aux = parse_urn(json_vtl[CL])
                ref = Reference(
                    sdmx_type=CL,
                    agency=cl_ref_aux.agency,
                    id=cl_ref_aux.id,
                    version=cl_ref_aux.version,
                )
            else:
                cl_ref = json_vtl[CL][REF]
                ref = Reference(
                    sdmx_type=CL,
                    agency=cl_ref[AGENCY_ID],
                    id=cl_ref[ID],
                    version=cl_ref[VERSION],
                )
            json_vtl[CL_LOW] = self.codelists.get(str(ref), ref)
            del json_vtl[CL]
            json_vtl["codelist_alias"] = json_vtl.pop("alias")
        # Concept mapping
        if CON in json_vtl:
            if isinstance(json_vtl[CON], str):
                con_ref_aux = parse_urn(json_vtl[CON])
                item_ref = ItemReference(
                    sdmx_type=CON,
                    agency=con_ref_aux.agency,
                    id=con_ref_aux.id,
                    version=con_ref_aux.version,
                    item_id=con_ref_aux.item_id,  # type: ignore[union-attr]
                )
            else:
                con_ref = json_vtl[CON][REF]
                item_ref = ItemReference(
                    sdmx_type=CON,
                    agency=con_ref[AGENCY_ID],
                    id=con_ref[PAR_ID],
                    version=con_ref[PAR_VER],
                    item_id=con_ref[ID],
                )
            json_vtl[CON_LOW] = self.concepts.get(str(item_ref), item_ref)
            del json_vtl[CON]
            json_vtl["concept_alias"] = json_vtl.pop("alias")
        # Custom type
        _format_lower_key("VtlScalarType", json_vtl)
        _format_lower_key("DataType", json_vtl)
        _format_lower_key("NullValue", json_vtl)
        _format_lower_key("OutputFormat", json_vtl)
        _format_lower_key("VtlLiteralFormat", json_vtl)

        # Name Personalisation
        _format_lower_key("PersonalisedName", json_vtl)
        _format_lower_key("vtlArtefact", json_vtl)
        _format_lower_key("VtlDefaultName", json_vtl)

        return json_vtl

    def __format_item(
        self, item_json_info: Dict[str, Any], item_name_class: str
    ) -> Item:
        item_json_info = self.__format_annotations(item_json_info)
        item_json_info = self.__format_name_description(item_json_info)
        if CONTACT in item_json_info and item_name_class == AGENCY:
            item_json_info[CONTACT] = add_list(item_json_info[CONTACT])
            contacts = []
            for e in item_json_info[CONTACT]:
                contacts.append(self.__format_contact(e))
            item_json_info[CONTACT.lower() + "s"] = contacts
            del item_json_info[CONTACT]

        if CORE_REP in item_json_info and item_name_class == CON:
            self.__format_representation(
                item_json_info[CORE_REP], item_json_info
            )
            del item_json_info[CORE_REP]

        if "Parent" in item_json_info:
            del item_json_info["Parent"]
        if DFW in item_json_info:
            self.__format_dataflow(item_json_info[DFW], item_json_info)

        item_json_info = self.__format_vtl(item_json_info)

        if CL_LOW in item_json_info and item_name_class == VTLMAPPING:
            item_name_class = VTL_CL_MAPP
        elif CON_LOW in item_json_info and item_name_class == VTLMAPPING:
            item_name_class = VTL_CON_MAPP

        return ITEMS_CLASSES[item_name_class](**item_json_info)

    @staticmethod
    def __format_groups(element: Dict[str, Any]) -> Dict[str, Any]:
        if DSD_COMPS in element:
            dsd_comps = element[DSD_COMPS]
            if GROUP in dsd_comps:
                groups = (
                    dsd_comps[GROUP]
                    if isinstance(dsd_comps[GROUP], list)
                    else [dsd_comps[GROUP]]
                )
                for group in groups:
                    group_dimensions = group.pop(GROUP_DIM, [])
                    if isinstance(group_dimensions, dict):
                        group_dimensions = [group_dimensions]

                    group["dimensions"] = [
                        d[DIM_REF]
                        if isinstance(d[DIM_REF], str)
                        else d[DIM_REF][REF][ID]
                        for d in group_dimensions
                    ]

                element[GROUPS_LOW] = [Group(**g) for g in groups]
                del element[DSD_COMPS][GROUP]
        return element

    def __format_is_final_30(
        self, json_elem: Dict[str, Any]
    ) -> Dict[str, Any]:
        if self.is_sdmx_30:
            # Default version value is 1.0, in SDMX-ML 3.0 we need to set
            # is_final as True if the version does not have an EXTENSION
            # (see Technical Notes SDMX 3.0)
            json_elem[IS_FINAL_LOW] = (
                "-" not in json_elem[VERSION] if VERSION in json_elem else True
            )
        return json_elem

    def __format_scheme(
        self, json_elem: Dict[str, Any], scheme: str, item: str
    ) -> Dict[str, ItemScheme]:
        elements: Dict[str, ItemScheme] = {}

        json_elem[scheme] = add_list(json_elem[scheme])
        for element in json_elem[scheme]:
            element["items"] = []

            element = self.__format_annotations(element)
            element = self.__format_name_description(element)
            element = self.__format_urls(element)
            if IS_EXTERNAL_REF in element:
                element[IS_EXTERNAL_REF_LOW] = (
                    element.pop(IS_EXTERNAL_REF) == "true"
                )
            if IS_FINAL in element:
                element[IS_FINAL_LOW] = element.pop(IS_FINAL) == "true"
            if IS_PARTIAL in element:
                element[IS_PARTIAL_LOW] = element.pop(IS_PARTIAL) == "true"
            items = []
            if item in element:
                element[item] = add_list(element[item])
                for item_elem in element[item]:
                    # Dynamic
                    items.append(self.__format_item(item_elem, item))
                del element[item]
            element["items"] = items
            element = self.__format_agency(element)
            element = self.__format_validity(element)
            element = self.__format_vtl(element)
            element = self.__format_vtl_references(element)
            if "xmlns" in element:
                del element["xmlns"]
            # Dynamic creation with specific class
            if scheme == VALUE_LIST:
                element["sdmx_type"] = VALUE_LIST_LOW
            element = self.__format_is_final_30(element)
            result: ItemScheme = STRUCTURES_MAPPING[scheme](**element)
            elements[result.short_urn] = result

        return elements

    def __format_schema(  # noqa: C901
        self, json_element: Dict[str, Any], schema: str, item: str
    ) -> Dict[str, Any]:
        """Formats the structures in json format.

        Args:
            json_element: The structures in json format
            schema: The scheme of the structures
            item: The item of the structures

        Returns:
            A dictionary with the structures formatted
        """
        schemas = {}

        json_element[item] = add_list(json_element[item])
        for element in json_element[item]:
            if URN.lower() in element and element[URN.lower()] is not None:
                short_urn = parse_urn(element[URN.lower()]).__str__()
            else:
                short_urn = Reference(
                    sdmx_type=item,
                    agency=element[AGENCY_ID],
                    id=element[ID],
                    version=element[VERSION],
                ).__str__()
            if METADATA in element:
                del element[METADATA]
            element = self.__format_annotations(element)
            element = self.__format_name_description(element)
            element = self.__format_urls(element)
            element = self.__format_agency(element)
            element = self.__format_validity(element)
            element = self.__format_groups(element)
            element = self.__format_components(element)

            if "xmlns" in element:
                del element["xmlns"]
            if IS_EXTERNAL_REF in element:
                element[IS_EXTERNAL_REF_LOW] = element.pop(IS_EXTERNAL_REF)
                element[IS_EXTERNAL_REF_LOW] = (
                    str(element[IS_EXTERNAL_REF_LOW]).lower() == "true"
                )
            if IS_FINAL in element:
                element[IS_FINAL_LOW] = element.pop(IS_FINAL)
                element[IS_FINAL_LOW] = (
                    str(element[IS_FINAL_LOW]).lower() == "true"
                )

            if item == DFW:
                if isinstance(element[STRUCTURE], str):
                    ref_obj = parse_urn(element[STRUCTURE])
                    reference_str = (
                        f"{ref_obj.sdmx_type}={ref_obj.agency}:"
                        f"{ref_obj.id}({ref_obj.version})"
                    )
                else:
                    ref_data = element[STRUCTURE][REF]
                    reference_str = (
                        f"{ref_data[CLASS]}={ref_data[AGENCY_ID]}"
                        f":{ref_data[ID]}({ref_data[VERSION]})"
                    )
                element[STRUCTURE] = reference_str

            structure = {key.lower(): value for key, value in element.items()}
            if schema == DSDS:
                if COMPS in structure:
                    structure[COMPS] = Components(structure[COMPS])
                else:
                    structure[COMPS] = Components([])
            structure = self.__format_is_final_30(structure)
            schemas[short_urn] = STRUCTURES_MAPPING[schema](**structure)

        return schemas

    def format_structures(
        self, json_meta: Dict[str, Any]
    ) -> Sequence[Union[ItemScheme, DataStructureDefinition, Dataflow]]:
        """Formats the structures in JSON format.

        Args:
            json_meta: The structures in JSON format.

        Returns:
            A list with the formatted structures.
        """

        def process_structure(
            key: str,
            formatter: Callable[[Dict[str, Any]], Dict[Any, Any]],
            attr: Optional[str] = None,
        ) -> Dict[Any, Any]:
            """Helper function to process and store formatted structures."""
            if key in json_meta:
                formatted = formatter(json_meta[key])
                setattr(self, attr, formatted) if attr else None
                return formatted
            return {}

        structures = {
            ORGS: process_structure(ORGS, self.__format_orgs, "agencies"),
            AGENCIES: process_structure(
                AGENCIES, self.__format_orgs, "agencies"
            ),
            CLS: process_structure(
                CLS,
                lambda data: self.__format_scheme(data, CL, CODE),
                "codelists",
            ),
            VALUE_LISTS: process_structure(
                VALUE_LISTS,
                lambda data: self.__format_scheme(
                    data, VALUE_LIST, VALUE_ITEM
                ),
                "valuelists",
            ),
            CON_SCHEMES: process_structure(
                CON_SCHEMES,
                lambda data: self.__format_scheme(data, CS, CON),
                "concepts",
            ),
            CONCEPTS: process_structure(
                CONCEPTS,
                lambda data: self.__format_scheme(data, CS, CON),
                "concepts",
            ),
            DSDS: process_structure(
                DSDS,
                lambda data: self.__format_schema(data, DSDS, DSD),
                "datastructures",
            ),
            DFWS: process_structure(
                DFWS,
                lambda data: self.__format_schema(data, DFWS, DFW),
                "dataflows",
            ),
            VTLMAPPINGS: process_structure(
                VTLMAPPINGS,
                lambda data: self.__format_scheme(
                    data,
                    VTL_MAPPING_SCHEME,
                    VTLMAPPING,
                ),
                "vtl_mappings",
            ),
            VTLMAPPING_SCHEMES: process_structure(
                VTLMAPPING_SCHEMES,
                lambda data: self.__format_scheme(
                    data,
                    VTL_MAPPING_SCHEME,
                    VTLMAPPING,
                ),
                "vtl_mappings",
            ),
            RULESETS: process_structure(
                RULESETS,
                lambda data: self.__format_scheme(data, RULE_SCHEME, RULE),
                "rulesets",
            ),
            RULE_SCHEMES: process_structure(
                RULE_SCHEMES,
                lambda data: self.__format_scheme(data, RULE_SCHEME, RULE),
                "rulesets",
            ),
            UDOS: process_structure(
                UDOS,
                lambda data: self.__format_scheme(data, UDO_SCHEME, UDO),
                "udos",
            ),
            UDO_SCHEMES: process_structure(
                UDO_SCHEMES,
                lambda data: self.__format_scheme(data, UDO_SCHEME, UDO),
                "udos",
            ),
            NAME_PERS: process_structure(
                NAME_PERS,
                lambda data: self.__format_scheme(
                    data, NAME_PER_SCHEME, NAME_PER
                ),
                "name_personalisations",
            ),
            NAME_PER_SCHEMES: process_structure(
                NAME_PER_SCHEMES,
                lambda data: self.__format_scheme(
                    data, NAME_PER_SCHEME, NAME_PER
                ),
                "name_personalisations",
            ),
            CUSTOM_TYPES: process_structure(
                CUSTOM_TYPES,
                lambda data: self.__format_scheme(
                    data, CUSTOM_TYPE_SCHEME, CUSTOM_TYPE
                ),
                "custom_types",
            ),
            CUSTOM_TYPE_SCHEMES: process_structure(
                CUSTOM_TYPE_SCHEMES,
                lambda data: self.__format_scheme(
                    data, CUSTOM_TYPE_SCHEME, CUSTOM_TYPE
                ),
                "custom_types",
            ),
            TRANSFORMATIONS: process_structure(
                TRANSFORMATIONS,
                lambda data: self.__format_scheme(
                    data,
                    TRANS_SCHEME,
                    TRANSFORMATION,
                ),
                "transformations",
            ),
            TRANS_SCHEMES: process_structure(
                TRANS_SCHEMES,
                lambda data: self.__format_scheme(
                    data,
                    TRANS_SCHEME,
                    TRANSFORMATION,
                ),
                "transformations",
            ),
        }
        return [
            compound
            for value in structures.values()
            if value
            for compound in value.values()
        ]



"""Collection of SDMX-JSON schemas for dataflow queries."""

from typing import List, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.code import JsonCodelist, JsonValuelist
from pysdmx.io.json.sdmxjson2.messages.concept import JsonConceptScheme
from pysdmx.io.json.sdmxjson2.messages.core import MaintainableType
from pysdmx.io.json.sdmxjson2.messages.dsd import JsonDataStructure
from pysdmx.io.json.sdmxjson2.messages.provider import JsonDataProviderScheme
from pysdmx.model import (
    Agency,
    Components,
    Dataflow,
    DataflowInfo,
    DataProvider,
    DataStructureDefinition,
)
from pysdmx.util import parse_urn


class JsonDataflow(MaintainableType, frozen=True):
    """SDMX-JSON payload for a dataflow."""

    structure: str = ""

    def to_model(
        self,
        dsds: Sequence[JsonDataStructure] = (),
        concepts: Sequence[JsonConceptScheme] = (),
        valuelists: Sequence[JsonValuelist] = (),
        codelists: Sequence[JsonCodelist] = (),
    ) -> Dataflow:
        """Converts a FusionDataflow to a standard dataflow."""
        dsd: Optional[Union[DataStructureDefinition, str]] = None
        if len(dsds) > 0:
            ref = parse_urn(self.structure)
            m = [
                d
                for d in dsds
                if d.agency == ref.agency
                and d.id == ref.id
                and d.version == ref.version
            ]
            if len(m) == 1:
                dsd = m[0].to_model(concepts, codelists, valuelists, ())
        dsd = dsd if dsd is not None else self.structure
        return Dataflow(
            id=self.id,
            agency=self.agency,
            name=self.name,
            description=self.description,
            version=self.version,
            structure=dsd,
            annotations=[a.to_model() for a in self.annotations],
            is_external_reference=self.isExternalReference,
            valid_from=self.validFrom,
            valid_to=self.validTo,
        )


class JsonDataflows(Struct, frozen=True):
    """SDMX-JSON payload for the list of dataflows."""

    dataflows: Sequence[JsonDataflow]
    dataProviderSchemes: Sequence[JsonDataProviderScheme] = ()
    dataStructures: Sequence[JsonDataStructure] = ()
    conceptSchemes: Sequence[JsonConceptScheme] = ()
    valuelists: Sequence[JsonValuelist] = ()
    codelists: Sequence[JsonCodelist] = ()

    def __filter(
        self, df: JsonDataflow, agency: str, id_: str, version: str
    ) -> bool:
        if version != "+" and version != "latest":
            return (
                df.agency == agency and df.id == id_ and df.version == version
            )
        else:
            return df.agency == agency and df.id == id_

    def to_model(
        self, components: Components, agency: str, id_: str, version: str
    ) -> DataflowInfo:
        """Returns the requested dataflow details."""
        prvs: List[DataProvider] = []
        for dps in self.dataProviderSchemes:
            prvs.extend(dps.dataProviders)
        df = list(
            filter(
                lambda df: self.__filter(df, agency, id_, version),
                self.dataflows,
            )
        )[0]
        return DataflowInfo(
            id=df.id,
            components=components,
            agency=Agency(df.agency),
            name=df.name,
            description=df.description,
            version=df.version,
            providers=prvs,
            dsd_ref=df.structure,
        )

    def to_generic_model(self) -> Sequence[Dataflow]:
        """Returns the requested dataflows."""
        return [
            df.to_model(
                self.dataStructures,
                self.conceptSchemes,
                self.valuelists,
                self.codelists,
            )
            for df in self.dataflows
        ]


class JsonDataflowMessage(Struct, frozen=True):
    """SDMX-JSON payload for /dataflow queries (with details)."""

    data: JsonDataflows

    def to_model(
        self, components: Components, agency: str, id_: str, version: str
    ) -> DataflowInfo:
        """Returns the requested dataflow details."""
        return self.data.to_model(components, agency, id_, version)


class JsonDataflowsMessage(Struct, frozen=True):
    """SDMX-JSON payload for /dataflow queries."""

    data: JsonDataflows

    def to_model(self) -> Sequence[Dataflow]:
        """Returns the requested dataflows."""
        return self.data.to_generic_model()



"""Collection of SDMX-JSON schemas for categories and category schemes."""

from collections import defaultdict
from typing import Dict, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import (
    ItemSchemeType,
    MaintainableType,
    NameableType,
)
from pysdmx.io.json.sdmxjson2.messages.dataflow import JsonDataflow
from pysdmx.model import (
    Agency,
    Categorisation,
    Category,
    CategoryScheme,
    Dataflow,
    DataflowRef,
)
from pysdmx.util import find_by_urn


class JsonCategorisation(
    MaintainableType, frozen=True, rename={"agency": "agencyID"}
):
    """SDMX-JSON payload for a categorisation."""

    source: str = ""
    target: str = ""

    def to_model(self) -> Categorisation:
        """Converts a JsonCategorisation to a standard categorisation."""
        return Categorisation(
            id=self.id,
            agency=self.agency,
            version=self.version,
            source=self.source,
            target=self.target,
            name=self.name,
            description=self.description,
            is_external_reference=self.isExternalReference,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            annotations=[a.to_model() for a in self.annotations],
        )


class JsonCategory(NameableType, frozen=True):
    """SDMX-JSON payload for a category."""

    categories: Sequence["JsonCategory"] = ()

    def to_model(self) -> Category:
        """Converts a FusionCode to a standard code."""
        return Category(
            id=self.id,
            name=self.name,
            description=self.description,
            categories=[c.to_model() for c in self.categories],
            annotations=[a.to_model() for a in self.annotations],
        )


class JsonCategoryScheme(
    ItemSchemeType, frozen=True, rename={"agency": "agencyID"}
):
    """SDMX-JSON payload for a category scheme."""

    categories: Sequence[JsonCategory] = ()

    def to_model(self) -> CategoryScheme:
        """Converts a JsonCodelist to a standard codelist."""
        return CategoryScheme(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            items=[c.to_model() for c in self.categories],
            is_external_reference=self.isExternalReference,
            is_partial=self.isPartial,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            annotations=[a.to_model() for a in self.annotations],
        )


class JsonCategorySchemes(Struct, frozen=True):
    """SDMX-JSON payload for the list of category schemes."""

    categorySchemes: Sequence[JsonCategoryScheme]
    categorisations: Sequence[JsonCategorisation] = ()
    dataflows: Sequence[JsonDataflow] = ()


class JsonCategorySchemeMessage(Struct, frozen=True):
    """SDMX-JSON payload for /categoryscheme queries."""

    data: JsonCategorySchemes

    def __group_flows(self) -> defaultdict[str, list[Dataflow]]:
        out: defaultdict[str, list[Dataflow]] = defaultdict(list)
        for c in self.data.categorisations:
            d = find_by_urn(self.data.dataflows, c.source)
            src = c.target[c.target.find(")") + 2 :]
            out[src].append(d.to_model())
        return out

    def __add_flows(
        self, cat: Category, cni: str, cf: Dict[str, list[Dataflow]]
    ) -> None:
        if cat.categories:
            for c in cat.categories:
                self.__add_flows(c, f"{cni}.{c.id}", cf)
        if cni in cf:
            dfrefs = [
                DataflowRef(
                    (
                        df.agency.id
                        if isinstance(df.agency, Agency)
                        else df.agency
                    ),
                    df.id,
                    df.version,
                    df.name,
                )
                for df in cf[cni]
            ]
            cat.dataflows = dfrefs

    def to_model(self) -> CategoryScheme:
        """Returns the requested codelist."""
        cf = self.__group_flows()
        cs = self.data.categorySchemes[0].to_model()
        for c in cs:
            self.__add_flows(c, c.id, cf)
        return cs


class JsonCategorisations(Struct, frozen=True):
    """SDMX-JSON payload for the list of categorisations."""

    categorisations: Sequence[JsonCategorisation]


class JsonCategorisationMessage(Struct, frozen=True):
    """SDMX-JSON payload for /categorisation queries."""

    data: JsonCategorisations

    def to_model(self) -> Sequence[Categorisation]:
        """Returns the requested categorisations."""
        return [cat.to_model() for cat in self.data.categorisations]



"""Connector to SDMX-REST and GDS-REST services."""

from typing import NoReturn, Optional, Union

import httpx

from pysdmx import errors
from pysdmx.api.qb.availability import AvailabilityFormat, AvailabilityQuery
from pysdmx.api.qb.data import DataFormat, DataQuery
from pysdmx.api.qb.gds import GdsQuery
from pysdmx.api.qb.refmeta import (
    RefMetaByMetadataflowQuery,
    RefMetaByMetadatasetQuery,
    RefMetaByStructureQuery,
    RefMetaFormat,
)
from pysdmx.api.qb.registration import (
    RegistrationByContextQuery,
    RegistrationByIdQuery,
    RegistrationByProviderQuery,
    RegistryFormat,
)
from pysdmx.api.qb.schema import SchemaFormat, SchemaQuery
from pysdmx.api.qb.structure import StructureFormat, StructureQuery
from pysdmx.api.qb.util import ApiVersion
from pysdmx.io.format import GDS_FORMAT
from pysdmx.util._net_utils import map_httpx_errors


class _CoreRestService:
    """Abstract connector."""

    def __init__(
        self,
        api_endpoint: str,
        api_version: ApiVersion,
        data_format: DataFormat,
        structure_format: StructureFormat,
        schema_format: SchemaFormat,
        refmeta_format: RefMetaFormat,
        avail_format: AvailabilityFormat,
        registry_format: RegistryFormat,
        pem: Optional[str] = None,
        timeout: Optional[float] = 5.0,
    ):
        """Instantiate a connector to a SDMX-REST service."""
        self._api_endpoint = (
            api_endpoint[0:-1] if api_endpoint.endswith("/") else api_endpoint
        )

        self._api_version = api_version
        self._data_format = data_format
        self._structure_format = structure_format
        self._schema_format = schema_format
        self._refmeta_format = refmeta_format
        self._avail_format = avail_format
        self._registry_format = registry_format
        self._ssl_context = (
            httpx.create_ssl_context(
                verify=pem,
            )
            if pem
            else httpx.create_ssl_context()
        )
        self._headers = {
            "Accept-Encoding": "gzip, deflate",
        }
        self._timeout = timeout


class RestService(_CoreRestService):
    """Synchronous connector to SDMX-REST services.

    Args:
        api_endpoint: The entry point (URL) of the SDMX-REST service.
        api_version: The most recent version of the SDMX-REST specification
            supported by the service.
        data_format: The default format for data queries.
        structure_format: The default format for structure queries.
        schema_format: The default format for schema queries.
        refmeta_format: The default format for reference metadata queries.
        avail_format: The default format for availability queries.
        registry_format: The default format for registration queries.
        pem: In case the service uses SSL/TLS with self-signed certificate,
            this attribute should be used to pass the pem file with the
            list of trusted certicate authorities.
        timeout: The maximum number of seconds to wait before considering
            that a request timed out. Defaults to 5 seconds.
    """

    def __init__(
        self,
        api_endpoint: str,
        api_version: ApiVersion,
        data_format: DataFormat = DataFormat.SDMX_JSON_2_0_0,
        structure_format: StructureFormat = StructureFormat.SDMX_JSON_2_0_0,
        schema_format: SchemaFormat = SchemaFormat.SDMX_JSON_2_0_0_STRUCTURE,
        refmeta_format: RefMetaFormat = RefMetaFormat.SDMX_JSON_2_0_0,
        avail_format: AvailabilityFormat = AvailabilityFormat.SDMX_JSON_2_0_0,
        registry_format: RegistryFormat = RegistryFormat.FUSION_JSON,
        pem: Optional[str] = None,
        timeout: Optional[float] = 5.0,
    ):
        """Instantiate a connector to a SDMX-REST service."""
        super().__init__(
            api_endpoint,
            api_version,
            data_format,
            structure_format,
            schema_format,
            refmeta_format,
            avail_format,
            registry_format,
            pem,
            timeout,
        )

    def data(self, query: DataQuery) -> bytes:
        """Execute a data query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._data_format.value
        return self.__fetch(q, f)

    def structure(self, query: StructureQuery) -> bytes:
        """Execute a structure query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._structure_format.value
        return self.__fetch(q, f)

    def schema(self, query: SchemaQuery) -> bytes:
        """Execute a schema query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._schema_format.value
        return self.__fetch(q, f)

    def availability(self, query: AvailabilityQuery) -> bytes:
        """Execute an availability query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._avail_format.value
        return self.__fetch(q, f)

    def reference_metadata(
        self,
        query: Union[
            RefMetaByMetadataflowQuery,
            RefMetaByMetadatasetQuery,
            RefMetaByStructureQuery,
        ],
    ) -> bytes:
        """Execute a reference metadata query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._refmeta_format.value
        return self.__fetch(q, f)

    def registration(
        self,
        query: Union[
            RegistrationByContextQuery,
            RegistrationByIdQuery,
            RegistrationByProviderQuery,
        ],
    ) -> bytes:
        """Execute a registration query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._registry_format.value
        return self.__fetch(q, f)

    def __fetch(self, query: str, format: str) -> bytes:
        with httpx.Client(verify=self._ssl_context) as client:
            try:
                url = f"{self._api_endpoint}{query}"
                h = self._headers.copy()
                h["Accept"] = format
                r = client.get(url, headers=h, timeout=self._timeout)
                r.raise_for_status()
                return r.content
            except (httpx.RequestError, httpx.HTTPStatusError) as e:
                map_httpx_errors(e)


class AsyncRestService(_CoreRestService):
    """Asynchronous connector to SDMX-REST services.

    Args:
        api_endpoint: The entry point (URL) of the SDMX-REST service.
        api_version: The most recent version of the SDMX-REST specification
            supported by the service.
        data_format: The default format for data queries.
        structure_format: The default format for structure queries.
        schema_format: The default format for schema queries.
        refmeta_format: The default format for reference metadata queries.
        avail_format: The default format for availability queries.
        registry_format: The default format for registration queries.
        pem: In case the service uses SSL/TLS with self-signed certificate,
            this attribute should be used to pass the pem file with the
            list of trusted certicate authorities.
        timeout: The maximum number of seconds to wait before considering
            that a request timed out. Defaults to 5 seconds.
    """

    def __init__(
        self,
        api_endpoint: str,
        api_version: ApiVersion,
        data_format: DataFormat = DataFormat.SDMX_JSON_2_0_0,
        structure_format: StructureFormat = StructureFormat.SDMX_JSON_2_0_0,
        schema_format: SchemaFormat = SchemaFormat.SDMX_JSON_2_0_0_STRUCTURE,
        refmeta_format: RefMetaFormat = RefMetaFormat.SDMX_JSON_2_0_0,
        avail_format: AvailabilityFormat = AvailabilityFormat.SDMX_JSON_2_0_0,
        registry_format: RegistryFormat = RegistryFormat.FUSION_JSON,
        pem: Optional[str] = None,
        timeout: Optional[float] = 5.0,
    ):
        """Instantiate a connector to a SDMX-REST service."""
        super().__init__(
            api_endpoint,
            api_version,
            data_format,
            structure_format,
            schema_format,
            refmeta_format,
            avail_format,
            registry_format,
            pem,
            timeout,
        )

    async def data(self, query: DataQuery) -> bytes:
        """Execute a data query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._data_format.value
        out = await self.__fetch(q, f)
        return out

    async def structure(self, query: StructureQuery) -> bytes:
        """Execute a structure query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._structure_format.value
        out = await self.__fetch(q, f)
        return out

    async def schema(self, query: SchemaQuery) -> bytes:
        """Execute a schema query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._schema_format.value
        out = await self.__fetch(q, f)
        return out

    async def availability(self, query: AvailabilityQuery) -> bytes:
        """Execute an availability query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._avail_format.value
        out = await self.__fetch(q, f)
        return out

    async def reference_metadata(
        self,
        query: Union[
            RefMetaByMetadataflowQuery,
            RefMetaByMetadatasetQuery,
            RefMetaByStructureQuery,
        ],
    ) -> bytes:
        """Execute a reference metadata query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._refmeta_format.value
        out = await self.__fetch(q, f)
        return out

    async def registration(
        self,
        query: Union[
            RegistrationByContextQuery,
            RegistrationByIdQuery,
            RegistrationByProviderQuery,
        ],
    ) -> bytes:
        """Execute a registration query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._registry_format.value
        out = await self.__fetch(q, f)
        return out

    async def __fetch(self, query: str, format: str) -> bytes:
        async with httpx.AsyncClient(verify=self._ssl_context) as client:
            try:
                url = f"{self._api_endpoint}{query}"
                h = self._headers.copy()
                h["Accept"] = format
                r = await client.get(url, headers=h, timeout=self._timeout)
                r.raise_for_status()
                return r.content
            except (httpx.RequestError, httpx.HTTPStatusError) as e:
                map_httpx_errors(e)


class _CoreGdsRestService:
    """Base class for GDS-REST services."""

    def __init__(
        self,
        api_endpoint: str,
        pem: Optional[str] = None,
        timeout: float = 5.0,
    ):
        self._api_endpoint = api_endpoint.rstrip("/")
        self._ssl_context = (
            httpx.create_ssl_context(verify=pem)
            if pem
            else httpx.create_ssl_context()
        )
        self._headers = {"Accept-Encoding": "gzip, deflate"}
        self._timeout = timeout

    def _map_error(
        self, e: Union[httpx.RequestError, httpx.HTTPStatusError]
    ) -> NoReturn:
        q = e.request.url
        if isinstance(e, httpx.HTTPStatusError):
            s = e.response.status_code
            t = e.response.text
            if s == 404:
                msg = (
                    f"The requested resource(s) could "
                    f"not be found. Query: `{q}`"
                )
                raise errors.NotFound("Not found", msg) from e
            elif s < 500:
                msg = f"Client error {s}. Query: `{q}`. Error: `{t}`."
                raise errors.Invalid(f"Client error {s}", msg) from e
            else:
                msg = f"Service error {s}. Query: `{q}`. Error: `{t}`."
                raise errors.InternalError(f"Service error {s}", msg) from e
        else:
            msg = f"Connection error. Query: `{q}`. Error: `{e}`."
            raise errors.Unavailable("Connection error", msg) from e


class GdsRestService(_CoreGdsRestService):
    """Synchronous GDS-REST service."""

    def _fetch(self, query: str, format_: str) -> bytes:
        with httpx.Client(verify=self._ssl_context) as client:
            try:
                url = f"{self._api_endpoint}{query}"
                headers = {**self._headers, "Accept": format_}
                response = client.get(
                    url, headers=headers, timeout=self._timeout
                )
                response.raise_for_status()
                return response.content
            except (httpx.RequestError, httpx.HTTPStatusError) as e:
                self._map_error(e)

    def gds(self, query: GdsQuery) -> bytes:
        """Execute a GDS query against the service."""
        q = query.get_url()
        return self._fetch(q, GDS_FORMAT)


class GdsAsyncRestService(_CoreGdsRestService):
    """Asynchronous GDS-REST service."""

    async def _fetch(self, query: str, format_: str) -> bytes:
        async with httpx.AsyncClient(verify=self._ssl_context) as client:
            try:
                url = f"{self._api_endpoint}{query}"
                headers = {**self._headers, "Accept": format_}
                response = await client.get(
                    url, headers=headers, timeout=self._timeout
                )
                response.raise_for_status()
                return response.content
            except (httpx.RequestError, httpx.HTTPStatusError) as e:
                self._map_error(e)

    async def gds(self, query: GdsQuery) -> bytes:
        """Execute a GDS query against the service."""
        q = query.get_url()
        return await self._fetch(q, GDS_FORMAT)



from pysdmx.errors import Invalid, NotFound
from pysdmx.model import Reference
from pysdmx.model.dataflow import Schema
from pysdmx.model.message import Message
from pysdmx.util import parse_urn


def schema_generator(message: Message, dataset_ref: Reference) -> Schema:
    """Generates a Schema by resolving the short_urn in the message."""
    context = dataset_ref.sdmx_type.lower()
    if context == "datastructure":
        try:
            dsd = message.get_data_structure_definition(str(dataset_ref))
        except NotFound:
            raise Invalid(
                f"Missing DataStructure {dataset_ref} in structures message.",
            ) from None
        return dsd.to_schema()
    elif context == "dataflow":
        try:
            dataflow = message.get_dataflow(str(dataset_ref))
        except NotFound:
            raise Invalid(
                f"Missing Dataflow {dataset_ref} in structures message.",
            ) from None
        if dataflow.structure is None:
            raise Invalid(
                f"Dataflow {dataset_ref} does not have a structure defined.",
            )
        dsd_ref = parse_urn(dataflow.structure)  # type: ignore[arg-type]
        try:
            dsd = message.get_data_structure_definition(str(dsd_ref))
        except NotFound:
            raise Invalid(
                f"Not found referenced DataStructure {dsd_ref}"
                f"from Dataflow {dataset_ref}. "
                f"Please send the structures message using "
                f"references=children to include the DataStructureDefinition.",
            ) from None
        return Schema(
            context=context,  # type: ignore[arg-type]
            id=dataset_ref.id,
            version=dataset_ref.version,
            agency=dataset_ref.agency,
            groups=dsd.groups,
            components=dsd.components,
            artefacts=dsd.to_schema().artefacts,
        )
    elif context == "provisionagreement":
        raise NotImplementedError(
            "ProvisionAgreement schema generation is not supported yet."
        )
    else:
        raise Invalid(
            f"Unknown context: {context}",
            f"The dataset can only reference a DataStructureDefinition, "
            f"a Dataflow or a Provision Agreement, "
            f"found {dataset_ref.sdmx_type}.",
        )



"""Collection of GDS-JSON schemas for GDS agencies."""

from typing import Sequence

from msgspec import Struct

from pysdmx.model import Agency


class JsonAgency(Struct, frozen=True):
    """GDS-JSON payload for an agency scheme."""

    agencyID: str
    name: str
    url: str
    description: str = ""

    def to_model(self) -> Agency:
        """Converts the payload to a GDS Agency."""
        return Agency(
            id=self.agencyID,
            name=self.name,
            uri=self.url,
            description=self.description,
        )


class JsonStructures(Struct, frozen=True):
    """Intermediate structure for 'structures' field."""

    agencies: Sequence[JsonAgency]


class JsonAgencyMessage(Struct, frozen=True):
    """GDS-JSON payload for /agency queries."""

    structures: JsonStructures

    def to_model(self) -> Sequence[Agency]:
        """Returns a list with the requested agencies."""
        return [a.to_model() for a in self.structures.agencies]



# mypy: disable-error-code="union-attr"
"""Module for writing SDMX-ML 3.0 Structure Specific data messages."""

from pathlib import Path
from typing import Dict, Optional, Sequence, Union

from pysdmx.io.format import Format
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.__tokens import (
    DSD_LOW,
    PROV_AGREEMENT,
    REGISTRY_LOW,
)
from pysdmx.io.xml.__write_aux import (
    __write_header,
    create_namespaces,
    get_end_message,
)
from pysdmx.io.xml.__write_data_aux import (
    check_content_dataset,
    check_dimension_at_observation,
)
from pysdmx.io.xml.__write_structure_specific_aux import (
    __write_data_structure_specific,
)
from pysdmx.model.message import Header
from pysdmx.util import parse_short_urn


def write(
    datasets: Sequence[PandasDataset],
    output_path: Optional[Union[str, Path]] = None,
    prettyprint: bool = True,
    header: Optional[Header] = None,
    dimension_at_observation: Optional[Dict[str, str]] = None,
) -> Optional[str]:
    """Write data to SDMX-ML 3.0 Structure Specific format.

    Args:
        datasets: The datasets to be written.
        output_path: The path to save the file.
        prettyprint: Prettyprint or not.
        header: The header to be used (generated if None).
        dimension_at_observation:
          The mapping between the dataset and the dimension at observation.

    Returns:
        The XML string if path is empty, None otherwise.
    """
    ss_namespaces = ""
    type_ = Format.DATA_SDMX_ML_3_0

    # Checking if we have datasets,
    # we need to ensure we can write them correctly
    check_content_dataset(datasets)
    content = {dataset.short_urn: dataset for dataset in datasets}

    if header is None:
        header = Header()

    # Checking the dimension at observation mapping
    dim_mapping = check_dimension_at_observation(
        content, dimension_at_observation
    )
    header.structure = dim_mapping
    add_namespace_structure = True
    for i, (short_urn, dimension) in enumerate(header.structure.items()):
        reference = parse_short_urn(short_urn)
        pre_urn = (
            REGISTRY_LOW if reference.sdmx_type == PROV_AGREEMENT else DSD_LOW
        )
        ss_namespaces += (
            f'xmlns:ns{i + 1}="urn:sdmx:org.sdmx'
            f".infomodel.{pre_urn}.{short_urn}"
            f':ObsLevelDim:{dimension}" '
        )

    # Generating the initial tag with namespaces
    outfile = create_namespaces(type_, ss_namespaces, prettyprint)
    # Generating the header
    outfile += __write_header(
        header,
        prettyprint,
        add_namespace_structure,
        data_message=True,
        references_30=True,
    )
    # Writing the content
    outfile += __write_data_structure_specific(
        datasets=content,
        dim_mapping=dim_mapping,
        prettyprint=prettyprint,
        references_30=True,
    )

    outfile += get_end_message(type_, prettyprint)

    output_path = (
        str(output_path) if isinstance(output_path, Path) else output_path
    )

    if output_path is None or output_path == "":
        return outfile

    with open(output_path, "w", encoding="UTF-8", errors="replace") as f:
        f.write(outfile)

    return None



"""Utility functions to leverage the SDMX information model in Pandas."""

from typing import Dict, Iterable

from pysdmx.model import Component, DataType


def __get_pd_type(dt: DataType, required: bool) -> str:  # noqa: C901
    if dt == DataType.SHORT:
        return "int16" if required else "Int16"
    elif dt == DataType.INTEGER:
        return "int32" if required else "Int32"
    elif dt == DataType.LONG or dt == DataType.COUNT:
        return "int64" if required else "Int64"
    elif dt == DataType.BIG_INTEGER or dt == DataType.DECIMAL:
        return "object"
    elif dt == DataType.FLOAT:
        return "float32" if required else "Float32"
    elif dt == DataType.DOUBLE:
        return "float64" if required else "Float64"
    elif dt == DataType.BOOLEAN:
        return "bool" if required else "boolean"
    elif dt == DataType.MONTH:
        return "int8" if required else "Int8"
    elif dt == DataType.GREGORIAN_TIME_PERIOD:
        return "object"
    elif dt == DataType.DATE_TIME:
        return "datetime64[ns]"
    elif dt == DataType.YEAR:
        return "datetime64[Y]"
    elif dt == DataType.YEAR_MONTH:
        return "datetime64[M]"
    elif dt == DataType.DATE:
        return "datetime64[D]"
    else:
        return "string"


def to_pandas_type(comp: Component) -> str:
    """Determine the appropriate Pandas data type for the given component.

    For enumerated components, returns 'category' as the Pandas data type.
    For non-enumerated components, maps the SDMX data type to the corresponding
    Pandas data type, taking into account whether the component is required.

    Args:
        comp (Component):
            The SDMX component for which to determine the Pandas data type.

    Returns:
        The string representation of the corresponding Pandas data type.
        Possible return values include:

        - 'category' (for enumerated components)
        - Numeric types ('int16', 'Int16', 'float32', 'Float32', etc.)
        - 'object' (for complex numeric types and time periods)
        - Datetime types ('datetime64[ns]', 'datetime64[Y]', etc.)
        - 'string' (default for unhandled types)
        - 'bool' or 'boolean' (for boolean values)
    """
    if comp.enumeration:
        return "category"
    else:
        return __get_pd_type(comp.dtype, comp.required)


def to_pandas_schema(components: Iterable[Component]) -> Dict[str, str]:
    """Infer the schema of a Pandas Data Frame from a list of components.

    This function generates a dictionary mapping component IDs to their
    corresponding Pandas data types. The resulting dictionary can be used
    as input to the Pandas `astype` method to cast DataFrame columns
    to the desired types.

    Args:
        components (Iterable[Component]):
            A collection of SDMX components from which the schema for
            the Pandas DataFrame will be inferred.

    Returns:
        Dict[str, str]:
            A dictionary where keys are the component IDs (field names)
            and values are their corresponding Pandas data types.
    """
    return {c.id: to_pandas_type(c) for c in components}



"""Module for GDS-JSON GDS message schemas.

This module provides classes and structures for handling GDS-JSON
messages related to GDS (Global Discovery Service) metadata, such as
agency information.

Exports:
    JsonAgencyMessage: Represents the GDS-JSON payload for agency queries.
"""

from pysdmx.io.json.gds.messages.agencies import (
    JsonAgencyMessage,
)
from pysdmx.io.json.gds.messages.catalog import (
    JsonCatalogMessage,
)
from pysdmx.io.json.gds.messages.sdmx_api import (
    JsonSdmxApiMessage,
)
from pysdmx.io.json.gds.messages.services import (
    JsonServiceMessage,
)
from pysdmx.io.json.gds.messages.urn_resolver import (
    JsonUrnResolverMessage,
    JsonUrnResolverResult,
)

__all__ = [
    "JsonAgencyMessage",
    "JsonCatalogMessage",
    "JsonSdmxApiMessage",
    "JsonServiceMessage",
    "JsonUrnResolverMessage",
    "JsonUrnResolverResult",
]



"""Collection of utility functions."""

import re
from typing import Any, Sequence, Union

from pysdmx.errors import Invalid, NotFound
from pysdmx.model import Agency, ItemReference, Reference
from pysdmx.util._date_pattern_map import convert_dpm
from pysdmx.util._net_utils import map_httpx_errors

NF = "Not found"

maintainable_urn_pattern = re.compile(r"^.*\.(.*)=(.*):(.*)\((.*)\)$")
item_urn_pattern = re.compile(r"^.*\.(.*)=(.*):(.*)\((.*)\)\.(.*)$")
short_urn_pattern = re.compile(r"^(.*)=(.*):(.*)\((.*)\)$")
short_item_urn_pattern = re.compile(r"^(.*)=(.*):(.*)\((.*)\)\.(.*)$")


def parse_urn(urn: str) -> Union[ItemReference, Reference]:
    """Parses an SDMX urn and returns the details."""
    try:
        return parse_maintainable_urn(urn)
    except Invalid:
        try:
            return parse_item_urn(urn)
        except Invalid:
            try:
                return parse_short_urn(urn)
            except Invalid:
                try:
                    return parse_short_item_urn(urn)
                except Invalid:
                    raise Invalid(
                        NF, "{urn} does not match any known pattern"
                    ) from None


def parse_maintainable_urn(urn: str) -> Reference:
    """Parses an SDMX maintainable urn and returns the details."""
    m = re.match(maintainable_urn_pattern, urn)
    if m:
        return Reference(
            sdmx_type=m.group(1),
            agency=m.group(2),
            id=m.group(3),
            version=m.group(4),
        )
    else:
        raise Invalid(NF, f"{urn} does not match {maintainable_urn_pattern}")


def parse_item_urn(urn: str) -> ItemReference:
    """Parses an SDMX item urn and returns the details."""
    m = re.match(item_urn_pattern, urn)
    if m:
        return ItemReference(
            sdmx_type=m.group(1),
            agency=m.group(2),
            id=m.group(3),
            version=m.group(4),
            item_id=m.group(5),
        )
    else:
        raise Invalid(NF, f"{urn} does not match {item_urn_pattern}.")


def parse_short_urn(urn: str) -> Reference:
    """Parses an SDMX short urn and returns the details."""
    m = re.match(short_urn_pattern, urn)
    if m:
        return Reference(
            sdmx_type=m.group(1),
            agency=m.group(2),
            id=m.group(3),
            version=m.group(4),
        )
    else:
        raise Invalid(NF, f"{urn} does not match {short_urn_pattern}.")


def parse_short_item_urn(urn: str) -> ItemReference:
    """Parses an SDMX short item urn and returns the details."""
    m = re.match(short_item_urn_pattern, urn)
    if m:
        return ItemReference(
            sdmx_type=m.group(1),
            agency=m.group(2),
            id=m.group(3),
            version=m.group(4),
            item_id=m.group(5),
        )
    else:
        raise Invalid(NF, f"{urn} does not match {short_item_urn_pattern}.")


def find_by_urn(artefacts: Sequence[Any], urn: str) -> Any:
    """Returns the maintainable artefact matching the supplied urn."""
    r = parse_urn(urn)
    f = [
        a
        for a in artefacts
        if (
            (
                a.agency == r.agency
                or (
                    a.agency.id == r.agency
                    if isinstance(a.agency, Agency)
                    else False
                )
            )
            and a.id == r.id
            and a.version == r.version
        )
    ]
    if f:
        return f[0]
    else:
        urns = [f"{a.agency}:{a.id}({a.version})" for a in artefacts]
        raise NotFound(
            NF,
            (
                f"Could not find an artefact matching the following URN: "
                f"{urn}. The artefacts received were: {urns}."
            ),
        )


__all__ = [
    "convert_dpm",
    "find_by_urn",
    "parse_item_urn",
    "parse_maintainable_urn",
    "parse_urn",
    "parse_short_urn",
    "parse_short_item_urn",
    "ItemReference",
    "Reference",
    "map_httpx_errors",
]



"""Collection of GDS-JSON schemas for GDS services."""

from typing import List, Optional, Sequence

from msgspec import Struct

from pysdmx.model.gds import GdsEndpoint, GdsService


class JsonService(Struct, frozen=True):
    """GDS-JSON payload for a service scheme."""

    agencyID: str
    id: str
    name: str
    urn: str
    version: str
    base: str
    endpoints: List[GdsEndpoint]
    authentication: Optional[str] = None

    def to_model(self) -> GdsService:
        """Converts the payload to a GDS Service."""
        return GdsService(
            agency=self.agencyID,
            id=self.id,
            name=self.name,
            urn=self.urn,
            version=self.version,
            base=self.base,
            endpoints=self.endpoints,
            authentication=self.authentication,
        )


class JsonStructures(Struct, frozen=True):
    """Intermediate structure for 'structures' field."""

    services: Sequence[JsonService]


class JsonServiceMessage(Struct, frozen=True):
    """GDS-JSON payload for /service queries."""

    structures: JsonStructures

    def to_model(self) -> Sequence[GdsService]:
        """Returns a list of GdsService objects."""
        return [s.to_model() for s in self.structures.services]
